/*	$Id: pt1.c,v 1.7 2015/04/06 02:51:36 tsubai Exp $	*/

/*-
 * Copyright (c) 2013,2014 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/callout.h>
#include <sys/conf.h>
#include <sys/device.h>
#include <sys/ioctl.h>
#include <sys/kernel.h>
#include <sys/kthread.h>
#include <sys/malloc.h>
#include <sys/sysctl.h>

#include <machine/bus.h>

#include <dev/pci/pcivar.h>
#include <dev/pci/pcireg.h>
#include <dev/pci/pcidevs.h>

#include "pt1reg.h"
#include "ptio.h"

#if defined(__powerpc__) || defined(__powerpc64__)
#define PT_SAVE_KVA	/* Save KVA space. */
#endif

#define PT_STAT

#define NDESC 32			/* N of descriptors (max. 4095) */
#define CHANNEL_BUFSIZE	(16*1024*1024)	/* per tuner buffer size */
#define MINREAD 65536			/* minimum transfer size */
#define PRIPT PRI_KERNEL_RT		/* tsleep priority */

struct pt_tuner {
	u_char opened;
	u_char addr;
	u_char type;
	u_char number;
	struct pt_channel current;
	u_int packet_offset;
	u_int rp;
	u_int wp;
#if 0
	u_int dma_warn;
	u_int dma_error;
	u_int dma_overflow;
#endif
	u_int tuner_error;
	u_int tuner_error_prev;
	u_int used;
	u_char *fifo;
	struct pt_softc *sc;
};

struct pt_softc {
	struct device sc_dev;
	bus_space_tag_t sc_iot;
	bus_space_handle_t sc_ioh;
	bus_dma_tag_t sc_dmat;

	int sc_pages;
	kmutex_t sc_lock;
	struct lwp *sc_kthread;
	struct pt_tuner sc_tuner[NTUNER];
	u_char sc_cardtype;
	u_int sc_tuner_power;
	int sc_lnbcnt;
	struct callout sc_lnb_ch;
	u_int64_t sc_id;
	void *sc_dmabuf;
	paddr_t sc_dmabuf_pa;
	paddr_t sc_dmadesc_pa;
	u_int sc_ramaddr;
};

struct i2c_cmd {
	u_char len;
	u_char value[7];
};

int pt1_match(struct device *, struct cfdata *, void *);
void pt1_attach(struct device *, struct device *, void *);
static int pt_dma_init(struct pt_softc *);
static void pt_reset_dma(struct pt_softc *);
static int pt_dispatch(struct pt_softc *, u_int);
static void kick_reader(struct pt_softc *, struct pt_tuner *);
static void pt_kthread(void *);
static void pt_tuner_reset(struct pt_softc *);
static void pt_start_stream(struct pt_softc *, struct pt_tuner *);
static void pt_stop_stream(struct pt_softc *, struct pt_tuner *);
static void pt_enable_lnb(struct pt_softc *, struct pt_tuner *);
static void pt_disable_lnb(struct pt_softc *, struct pt_tuner *);
static void lnb_off(void *);
static u_int pt_tuner_error(struct pt_softc *, struct pt_tuner *);
static int pt_set_channel(struct pt_softc *, struct pt_tuner *,
    struct pt_channel *);
static void *dma_alloc(struct pt_softc *, long, paddr_t *);
static void pt_sysctl_init(struct pt_softc *);

static int init_isdb_s(struct pt_softc *, u_int);
static int init_isdb_t(struct pt_softc *, u_int);
static int pt_tuner_init(struct pt_softc *);
static void s_sleep(struct pt_softc *, int);
static void t_sleep(struct pt_softc *, int);
static void s_wakeup(struct pt_softc *, int);
static void t_wakeup(struct pt_softc *, int);
static void pt_tuner_sleep(struct pt_softc *, struct pt_tuner *);
static void pt_tuner_wakeup(struct pt_softc *, struct pt_tuner *);
static u_int s_frequency(u_int);
static u_int s_lookup_cb(u_int);
static int s_tune(struct pt_softc *, struct pt_tuner *, int);
static int read_tsid(struct pt_softc *, struct pt_tuner *);
static int ts_lock(struct pt_softc *, struct pt_tuner *, u_int);
static u_int t_frequency(u_int);
static u_int t_lookup_cb(u_int);
static int t_tune(struct pt_softc *, struct pt_tuner *, int);
static int s_cnr(struct pt_softc *, u_int);
static int t_cnr(struct pt_softc *, u_int);
static int s_agc(struct pt_softc *, u_int);
static int t_agc(struct pt_softc *, u_int);
static int s_carrier(struct pt_softc *, u_int);
static int t_carrier(struct pt_softc *, int);
static int s_clock(struct pt_softc *, u_int);
static int t_clock(struct pt_softc *, u_int);
static int s_raw_err(struct pt_softc *, u_int);
static int t_raw_err(struct pt_softc *, u_int);

static void pt_i2c(struct pt_softc *, u_int, u_int, u_int, u_int, u_int, u_int);
static u_int64_t xc3s_read_id(struct pt_softc *);
static int xc3s_unlock(struct pt_softc *);
static int xc3s_pci_reset(struct pt_softc *);
static int xc3s_ram_reset(struct pt_softc *);
static int xc3s_ram_shift(struct pt_softc *, u_int);
static int xc3s_init(struct pt_softc *);
static void writebit(struct pt_softc *, u_int);
static void readbit(struct pt_softc *);
static void sendbyte(struct pt_softc *, u_int);
static void i2c_init(struct pt_softc *);
static void i2c_begin(struct pt_softc *);
static void i2c_start(struct pt_softc *, int);
static void i2c_stop(struct pt_softc *);
static void i2c_prepare_write(struct pt_softc *, u_int, struct i2c_cmd *);
static void i2c_prepare_read(struct pt_softc *, u_int, struct i2c_cmd *, int);
static void i2c_write(struct pt_softc *, u_int, struct i2c_cmd *);
static u_int i2c_read(struct pt_softc *, u_int, struct i2c_cmd *, int);

CFATTACH_DECL(opt, sizeof(struct pt_softc), pt1_match, pt1_attach, NULL);

extern struct cfdriver opt_cd;

dev_type_open(pt1open);
dev_type_close(pt1close);
dev_type_read(pt1read);
dev_type_ioctl(pt1ioctl);
dev_type_mmap(pt1mmap);

const struct cdevsw opt_cdevsw = {
	pt1open, pt1close, pt1read, nullwrite, pt1ioctl,
	nostop, notty, nopoll, pt1mmap, nokqfilter
};

/* LNB voltage control */
int pt_lnb_mode = LNB_15V;

/* DMAСĹ4KBʤΤǡ(Ψ̤)4KBñ̤ǽ */
struct pt_dmabuf {
	volatile u_int32_t mpkt[1024];
};
#define BUFFER_FILLED(p) ((p)->mpkt[1023])

enum { CARD_TYPE_PT1, CARD_TYPE_PT2 };
enum { TUNER_TYPE_ISDB_S, TUNER_TYPE_ISDB_T };

#define PT_READ(sc, reg) bus_space_read_4((sc)->sc_iot, (sc)->sc_ioh, reg)
#define PT_WRITE(sc, reg, val) \
    bus_space_write_4((sc)->sc_iot, (sc)->sc_ioh, reg, val)

int
pt1_match(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{
	struct pci_attach_args *pa = aux;

	if (PCI_VENDOR(pa->pa_id) == PCI_VENDOR_XILINX) {
		switch (PCI_PRODUCT(pa->pa_id)) {
		case PCI_PRODUCT_PT1:
		case PCI_PRODUCT_PT2:
			return 1;
		}
	}
	return 0;
}

void
pt1_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct pt_softc *sc = device_private(self);
	struct pci_attach_args *pa = aux;
	struct pt_tuner *tuner;
	char *cardname;
	u_int csr, class;
	int ch;

	csr = pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_COMMAND_STATUS_REG);
	csr |= PCI_COMMAND_MASTER_ENABLE;
	pci_conf_write(pa->pa_pc, pa->pa_tag, PCI_COMMAND_STATUS_REG, csr);

	switch (PCI_PRODUCT(pa->pa_id)) {
	case PCI_PRODUCT_PT1:
		cardname = "PT1";
		sc->sc_cardtype = CARD_TYPE_PT1;
		break;
	case PCI_PRODUCT_PT2:
		cardname = "PT2";
		sc->sc_cardtype = CARD_TYPE_PT2;
		break;
	default:
		printf(": unknown hardware\n");
		return;
	}

	class = pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_CLASS_REG);
	if (PCI_REVISION(class) != 1) {
		printf(": unknown revision\n");
		return;
	}

	if (pci_mapreg_map(pa, 0x10, PCI_MAPREG_TYPE_MEM, 0,
	    &sc->sc_iot, &sc->sc_ioh, NULL, NULL)) {
		printf(": failed to map device\n");
		return;
	}

	sc->sc_dmat = pa->pa_dmat;

	if (xc3s_init(sc)) {
		printf("%s: failed to initialize FPGA\n", self->dv_xname);
		goto out;
	}

	mutex_init(&sc->sc_lock, MUTEX_DEFAULT, IPL_NONE);
	callout_init(&sc->sc_lnb_ch, 0);

	for (ch = 0; ch < NTUNER; ch++) {
		static struct {
			u_char addr;
			u_char type;
			u_char ch;
		} tuners[] = {
			{ TUNER_ADDR_S1, TUNER_TYPE_ISDB_S, 0 },
			{ TUNER_ADDR_T1, TUNER_TYPE_ISDB_T, 63 },
			{ TUNER_ADDR_S2, TUNER_TYPE_ISDB_S, 0 },
			{ TUNER_ADDR_T2, TUNER_TYPE_ISDB_T, 63 }
		};

		tuner = &sc->sc_tuner[ch];
#ifdef PT_SAVE_KVA
		paddr_t pa;
		tuner->fifo = dma_alloc(sc, CHANNEL_BUFSIZE, &pa);
#else
		tuner->fifo = malloc(CHANNEL_BUFSIZE, M_DEVBUF, M_NOWAIT);
#endif
		if (tuner->fifo == NULL) {
			printf(": failed to allocate channel memory\n");
			goto out;
		}
		tuner->sc = sc;
		tuner->addr = tuners[ch].addr;
		tuner->type = tuners[ch].type;
		tuner->number = ch;
		tuner->rp = 0;
		tuner->wp = 0;
		tuner->current.freq = tuners[ch].ch;
	}

	pt_tuner_reset(sc);

	if (pt_tuner_init(sc)) {
		printf("%s: failed to initialize tuners\n", self->dv_xname);
		goto out;
	}
	for (ch = 0; ch < NTUNER; ch++) {
		pt_tuner_sleep(sc, &sc->sc_tuner[ch]);
		delay(1000);
	}

	if (pt_dma_init(sc)) {
		printf(": failed to allocate DMA buffer\n");
		goto out;
	}

	printf(": %s rev. %d\n", cardname, PCI_REVISION(class));
	printf("%s: using %d KB/DMA, %d KB/tuner\n", self->dv_xname,
	    NDESC * 512 * (PT_PGSZ / 1024), CHANNEL_BUFSIZE / 1024);

	kthread_create(PRI_NONE, 0, NULL, pt_kthread, sc, &sc->sc_kthread,
	    self->dv_xname);

	pt_sysctl_init(sc);
	return;

out:
	PT_WRITE(sc, PT_COMMAND_REG, ON(PCI_RESET|RAM_RESET|STOP_DMA));
	PT_WRITE(sc, PT_TUNER_REG, 0);
}

int
pt_dma_init(sc)
	struct pt_softc *sc;
{
	int i, j;
	paddr_t pa, addr;
	struct pt_dmadesc *p;
	char *va;

	va = dma_alloc(sc, PT_PGSZ * NDESC * 512, &pa);
	if (va == NULL)
		return -1;

	sc->sc_dmadesc_pa = pa;
	sc->sc_dmabuf    = va + PT_PGSZ * NDESC;
	sc->sc_dmabuf_pa = pa + PT_PGSZ * NDESC;

	p = (void *)va;
	pa = sc->sc_dmabuf_pa;
	for (i = 0; i < NDESC; i++) {
		addr = sc->sc_dmadesc_pa + ((i + 1) % NDESC) * PT_PGSZ;
		p->next = htole32(addr >> 12);

		for (j = 0; j < 511; j++) {
			p->pa[j] = htole32(pa >> 12);
			pa += PT_PGSZ;
		}
		p++;
	}
	return 0;
}

static void
pt_reset_dma(sc)
	struct pt_softc *sc;
{
	struct pt_dmabuf *dp;
	int i;

	/* Initialize buffers. */
	dp = sc->sc_dmabuf;
	for (i = 0; i < NDESC * 511; i++) {
		BUFFER_FILLED(dp) = 0;
		dp++;
	}
	/* Reset transfer counter. */
	PT_WRITE(sc, PT_COMMAND_REG, RESET_DMACTR);

	/* Configure transfer counter. */
	for (i = 0; i < MIN(NDESC, 4095); i++)
		PT_WRITE(sc, PT_COMMAND_REG, INC_DMACTR);

	/* Set DMA descriptor address. */
	PT_WRITE(sc, PT_DMA_REG, sc->sc_dmadesc_pa >> 12);

	/* Start DMA. */
	PT_WRITE(sc, PT_COMMAND_REG, OFF(STOP_DMA|FAST_BTB_EN) | START_DMA);
}

#if 0
static void
pt_stop_dma(sc)
	struct pt_softc *sc;
{
	int i;

	PT_WRITE(sc, PT_COMMAND_REG, ON(STOP_DMA));
	for (i = 0; i < 1000; i++) {
		if ((PT_READ(sc, PT_STATUS_REG) & DMA_BUSY) == 0)
			break;
		delay(1000);
	}
}
#endif

static inline int
pt_dispatch(sc, mpkt)
	struct pt_softc *sc;
	u_int mpkt;
{
	struct pt_tuner *tuner;
	u_int ch, val, p;

	ch = MICROPACKET_CHANNEL(mpkt) - 1;
	if (ch >= NTUNER) {
		printf("%s: micropacket error (0x%08x)\n",
		    sc->sc_dev.dv_xname, mpkt);
		return 0;
	}
	tuner = &sc->sc_tuner[ch];

	if (mpkt & MICROPACKET_ERROR) {
		val = PT_READ(sc, PT_STATUS_REG);
		printf("%s: DMA error 0x%x\n", sc->sc_dev.dv_xname, val);
#if 0
		if (val & RAM_OVERFLOW)
			tuner->dma_overflow++;
		if (val & DMA_ERROR)
			tuner->dma_error++;
		if (val & DMA_WARN)
			tuner->dma_warn++;
#endif
		/* Reset and restart DMA. */
		return -1;
	}
	if (tuner->opened == 0)
		return 0;
	if (mpkt & MICROPACKET_START)
		tuner->packet_offset = 0;
	p = tuner->wp;
	tuner->fifo[p++] = MICROPACKET_BYTE0(mpkt);
	p %= CHANNEL_BUFSIZE;
	tuner->fifo[p++] = MICROPACKET_BYTE1(mpkt);
	p %= CHANNEL_BUFSIZE;
	if (tuner->packet_offset++ == TSPKTLEN / 3) {
		/* The last two bytes in a TS packet. */
		;
	} else {
		tuner->fifo[p++] = MICROPACKET_BYTE2(mpkt);
		p %= CHANNEL_BUFSIZE;
	}
	tuner->wp = p;
	return 0;
}

static inline void
kick_reader(sc, tuner)
	struct pt_softc *sc;
	struct pt_tuner *tuner;
{
	u_int n;

	if (tuner->opened == 0)
		return;
	n = (tuner->wp - tuner->rp) % CHANNEL_BUFSIZE;
	if (n > CHANNEL_BUFSIZE - PT_PGSZ*3/4) {
		printf("%s: tuner %d buffer full\n", sc->sc_dev.dv_xname,
		    tuner->number);
	}
	if (n >= MINREAD)
		wakeup(tuner);
	if (n > tuner->used)
		tuner->used = n;
}

static void
pt_kthread(v)
	void *v;
{
	struct pt_softc *sc = v;
	struct pt_dmabuf *dp;
	int i, j, k, timo;
	int pages = 0;

restart:
	pt_reset_dma(sc);
loop:
	dp = sc->sc_dmabuf;
	for (i = 0; i < NDESC; i++) {
		for (j = 0; j < 511; j++) {
			/* Wait for the buffer filled in. */
			while (BUFFER_FILLED(dp) == 0) {
				timo = 0;
				for (k = 0; k < NTUNER; k++)
					if (sc->sc_tuner[k].opened)
						timo = hz / 16;
				if (pages > sc->sc_pages)
					sc->sc_pages = pages;
				pages = 0;
				tsleep(&sc->sc_kthread, PRIPT, "ptwait", timo);
			}
			/* XXX ޤPT¦̤žΥǡ󤸤ʥ? */

			pages++;
			for (k = 0; k < 1024; k++) {
				if (pt_dispatch(sc, dp->mpkt[k]))
					goto restart;
			}

			/* Mark this page is free now. */
			BUFFER_FILLED(dp) = 0;
			dp++;

			/* Wake up readers. */
			for (k = 0; k < NTUNER; k++)
				kick_reader(sc, &sc->sc_tuner[k]);
		}

		/* Increment DMA transfer counter. */
		PT_WRITE(sc, PT_COMMAND_REG, INC_DMACTR);
	}
	goto loop;
}

void
pt_tuner_reset(sc)
	struct pt_softc *sc;
{
	sc->sc_tuner_power = PT_TUNER_POWER | PT_TUNER_RESET |
	    PT_TUNER_POWER_S1 | PT_TUNER_POWER_S2 |
	    PT_TUNER_POWER_T1 | PT_TUNER_POWER_T2;
	PT_WRITE(sc, PT_TUNER_REG, sc->sc_tuner_power & ~PT_TUNER_RESET);
	delay(20000);	/* 15ms (min.) */
	PT_WRITE(sc, PT_TUNER_REG, sc->sc_tuner_power);
	delay(1000);	/* 1us (min.) */
}

static inline void
pt_start_stream(sc, tuner)
	struct pt_softc *sc;
	struct pt_tuner *tuner;
{
	PT_WRITE(sc, PT_TEST_EN_REG, ON(1) << tuner->number);
	/* ON(1 << chan) */
}

static inline void
pt_stop_stream(sc, tuner)
	struct pt_softc *sc;
	struct pt_tuner *tuner;
{
	PT_WRITE(sc, PT_TEST_EN_REG, OFF(1 << tuner->number));
}

void
pt_enable_lnb(sc, tuner)
	struct pt_softc *sc;
	struct pt_tuner *tuner;
{
	u_int val = 0;

	if (tuner->type != TUNER_TYPE_ISDB_S)
		return;

	mutex_enter(&sc->sc_lock);
	if (sc->sc_lnbcnt == 0) {
		callout_stop(&sc->sc_lnb_ch);
		switch (pt_lnb_mode) {
		case LNB_11V:
			val = PT_LNB_11V;
			break;
		case LNB_15V:
			val = PT_LNB_15V;
			break;
		}
		PT_WRITE(sc, PT_TUNER_REG, val | sc->sc_tuner_power);
		/* printf("%s: LNB on\n", sc->sc_dev.dv_xname); */

		tsleep(sc, 0, "lnbwait", hz/2);
	}
	sc->sc_lnbcnt++;
	mutex_exit(&sc->sc_lock);
}

void
pt_disable_lnb(sc, tuner)
	struct pt_softc *sc;
	struct pt_tuner *tuner;
{
	if (tuner->type != TUNER_TYPE_ISDB_S)
		return;

	mutex_enter(&sc->sc_lock);
	sc->sc_lnbcnt--;
	if (sc->sc_lnbcnt == 0)
		callout_reset(&sc->sc_lnb_ch, hz*10, lnb_off, sc);
	mutex_exit(&sc->sc_lock);
}

void
lnb_off(v)
	void *v;
{
	struct pt_softc *sc = v;

	PT_WRITE(sc, PT_TUNER_REG, sc->sc_tuner_power);
	/* printf("%s: LNB off\n", sc->sc_dev.dv_xname); */
}

u_int
pt_tuner_error(sc, tuner)
	struct pt_softc *sc;
	struct pt_tuner *tuner;
{
	u_int gray, val;
	int i;

	gray = PT_READ(sc, PT_ERROR_S1 + tuner->number * 4);
	gray &= 0x00ffffff;
	for (val = 0, i = 23; i >= 0; i--)
		val |= (gray ^ (val >> 1)) & (1 << i);
	return val;
}

static inline int
pt_set_channel(sc, tuner, ch)
	struct pt_softc *sc;
	struct pt_tuner *tuner;
	struct pt_channel *ch;
{
	int tsid;

	switch (tuner->type) {
	case TUNER_TYPE_ISDB_S:
		if (ch->freq < 0 || ch->freq > 35)
			return EINVAL;
		if (ch->slot < 0 || ch->slot > 7)
			return EINVAL;

		tuner->current = *ch;
		if (s_tune(sc, tuner, 1) < 0)
			return EIO;
		tsid = read_tsid(sc, tuner);
		if (tsid < 0)
			return EIO;
		if (tsid == 0xffff)
			return EINVAL;
		if (ts_lock(sc, tuner, tsid) < 0)
			return EIO;
		break;
	case TUNER_TYPE_ISDB_T:
		if (ch->freq < 0 || ch->freq > 112)
			return EINVAL;
		if (abs(ch->slot) > 28)
			return EINVAL;

		tuner->current = *ch;
		if (t_tune(sc, tuner, 1) < 0)
			return EIO;
	}
	return 0;
}

void *
dma_alloc(sc, size, pap)
	struct pt_softc *sc;
	long size;
	paddr_t *pap;
{
	bus_dma_tag_t t = sc->sc_dmat;
	bus_dmamap_t map;
	bus_dma_segment_t segs[1];
	void *kva;
	int n;

	/* Allocate and map a contiguous memory region. */
	if (bus_dmamem_alloc(t, size, PT_PGSZ, 0, segs, 1, &n, BUS_DMA_NOWAIT))
		goto out1;
#ifdef PT_SAVE_KVA
	*pap = segs[0].ds_addr;
	kva = (void *)segs[0].ds_addr;
	bzero(kva, size);
	return kva;
#endif
	if (bus_dmamem_map(t, segs, 1, size, &kva, BUS_DMA_NOWAIT))
		goto out2;
	if (bus_dmamap_create(t, size, 1, size, 0, BUS_DMA_NOWAIT, &map))
		goto out3;
	if (bus_dmamap_load(t, map, kva, size, NULL, BUS_DMA_NOWAIT))
		goto out4;

	bzero(kva, size);
	*pap = map->dm_segs[0].ds_addr;
	bus_dmamap_unload(t, map);
	bus_dmamap_destroy(t, map);
	return kva;

out4:
	bus_dmamap_destroy(t, map);
out3:
	bus_dmamem_unmap(t, kva, size);
out2:
	bus_dmamem_free(t, segs, 1);
out1:
	return NULL;
}

#define PTUNIT(x) (minor(x) >> 2)
#define PTCHAN(x) (minor(x) & 3)

int
pt1open(dev, flag, mode, l)
	dev_t dev;
	int flag, mode;
	struct lwp *l;
{
	struct pt_softc *sc;
	struct pt_tuner *tuner;

	sc = device_lookup_private(&opt_cd, PTUNIT(dev));
	if (sc == NULL)
		return ENXIO;

	tuner = &sc->sc_tuner[PTCHAN(dev)];
	if (tuner->opened)
		return EBUSY;
	tuner->opened = 1;
	tuner->packet_offset = 0;
	tuner->rp = tuner->wp = 0;
#if 0
	tuner->dma_warn = 0;
	tuner->dma_error = 0;
	tuner->dma_overflow = 0;
#endif

	pt_enable_lnb(sc, tuner);
	pt_tuner_wakeup(sc, tuner);
	tsleep(sc, 0, "ptopen", hz/8);
	return 0;
}

int
pt1close(dev, flag, mode, l)
	dev_t dev;
	int flag, mode;
	struct lwp *l;
{
	struct pt_softc *sc;
	struct pt_tuner *tuner;

	sc = device_lookup_private(&opt_cd, PTUNIT(dev));
	if (sc == NULL)
		return ENXIO;

	tuner = &sc->sc_tuner[PTCHAN(dev)];
	pt_stop_stream(sc, tuner);

	/* Turn off the tuner. */
	pt_tuner_sleep(sc, tuner);
	tsleep(sc, 0, "ptclose", hz/2);

	pt_disable_lnb(sc, tuner);

	tuner->rp = tuner->wp;
	tuner->opened = 0;
	return 0;
}

int
pt1read(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	struct pt_softc *sc;
	struct pt_tuner *tuner;
	u_int len, n;
	int error = 0;

	sc = device_lookup_private(&opt_cd, PTUNIT(dev));
	if (sc == NULL)
		return ENXIO;
	tuner = &sc->sc_tuner[PTCHAN(dev)];

	/* Sleep until we have some amount of data. */
	n = (tuner->wp - tuner->rp) % CHANNEL_BUFSIZE;
	if (n < MINREAD) {
		tsleep(tuner, PRIPT|PCATCH, "ptread", 0);
		n = (tuner->wp - tuner->rp) % CHANNEL_BUFSIZE;
	}
	len = MIN(uio->uio_resid, n & ~7);

	if (tuner->rp + len >= CHANNEL_BUFSIZE) {
		u_int rem = CHANNEL_BUFSIZE - tuner->rp;

		error = uiomove(tuner->fifo + tuner->rp, rem, uio);
		if (error)
			goto out;
		tuner->rp = 0;
		len -= rem;
	}
	error = uiomove(tuner->fifo + tuner->rp, len, uio);
	if (error)
		goto out;
	tuner->rp += len;

out:
	return error;
}

int
pt1ioctl(dev, cmd, data, flag, l)
	dev_t dev;
	u_long cmd;
	void *data;
	int flag;
	struct lwp *l;
{
	struct pt_softc *sc;
	struct pt_tuner *tuner;
	int error, i, *ip = data;

	sc = device_lookup_private(&opt_cd, PTUNIT(dev));
	if (sc == NULL)
		return ENXIO;
	tuner = &sc->sc_tuner[PTCHAN(dev)];

	switch(cmd) {
	case PTIOSCHAN:
	case SET_CHANNEL:
		/* XXX To be safe. */
		for (i = 0; i < 3; i++) {
			error = pt_set_channel(sc, tuner, data);
			if (error == 0)
				break;
		}

		/* Wait for the tuner stable. */
		tuner->tuner_error_prev = pt_tuner_error(sc, tuner);
		for (i = 0; i < 30; i++) {
			int x;

			tsleep(sc, 0, "ptioctl", hz/16);
			x = pt_tuner_error(sc, tuner);
			if (x == tuner->tuner_error_prev)
				break;
			tuner->tuner_error_prev = x;
		}
		return error;

	case PTIOGCHAN:
		*(struct pt_channel *)data = tuner->current;
		return 0;

	case PTIORECORD:
		if (*ip)
			goto start_rec;
		else
			goto stop_rec;

	case START_REC:
	start_rec:
		/* Clear tuner error counter. */
		tuner->tuner_error_prev = pt_tuner_error(sc, tuner);
		pt_start_stream(sc, tuner);
		wakeup(&sc->sc_kthread);
		return 0;

	case STOP_REC:
	stop_rec:
		pt_stop_stream(sc, tuner);
		tsleep(sc, 0, "ptioctl", hz/2);
		return 0;

	case PTIOGLNBMODE:
		*ip = pt_lnb_mode;
		return 0;

	case PTIOSLNBMODE:
		switch (*ip) {
		case LNB_OFF:
		case LNB_11V:
		case LNB_15V:
			pt_lnb_mode = *ip;
			return 0;
		}
		return EINVAL;

	case PTIOGCNR:
	case GET_SIGNAL_STRENGTH:
		switch(tuner->type) {
		case TUNER_TYPE_ISDB_S:
			*ip = s_cnr(sc, tuner->addr);
			return 0;
		case TUNER_TYPE_ISDB_T:
			*ip = t_cnr(sc, tuner->addr);
			return 0;
		}
		break;

	case PTIOGAGC:
		switch (tuner->type) {
		case TUNER_TYPE_ISDB_S:
			*ip = s_agc(sc, tuner->addr);
			return 0;
		case TUNER_TYPE_ISDB_T:
			*ip = t_agc(sc, tuner->addr);
			return 0;
		}
		break;

	case PTIOGCAR:
		switch (tuner->type) {
		case TUNER_TYPE_ISDB_S:
			*ip = s_carrier(sc, tuner->addr);
			return 0;
		case TUNER_TYPE_ISDB_T:
			*ip = t_carrier(sc, tuner->addr);
			return 0;
		}
		break;

	case PTIOGCLK:
		switch (tuner->type) {
		case TUNER_TYPE_ISDB_S:
			*ip = s_clock(sc, tuner->addr);
			return 0;
		case TUNER_TYPE_ISDB_T:
			*ip = t_clock(sc, tuner->addr);
			return 0;
		}
		break;

	case PTIOGRAWERR:
		switch (tuner->type) {
		case TUNER_TYPE_ISDB_S:
			*ip = s_raw_err(sc, tuner->addr);
			return 0;
		case TUNER_TYPE_ISDB_T:
			*ip = t_raw_err(sc, tuner->addr);
			return 0;
		}
		break;

	case LNB_ENABLE:
	case LNB_DISABLE:
		/* nothing to do */
		return 0;
	}
	return EINVAL;
}

paddr_t
pt1mmap(dev, off, prot)
	dev_t dev;
	off_t off;
	int prot;
{
	struct pt_softc *sc;

	sc = device_lookup_private(&opt_cd, PTUNIT(dev));
	if (sc == NULL)
		return ENXIO;

	if (off < 0)
		return -1;
	if (off >= (off_t)NDESC * 511 * PT_PGSZ)
		return -1;
	return sc->sc_dmabuf_pa + off;
}

static int
sysctl_pt_error(SYSCTLFN_ARGS)
{
	struct sysctlnode node = *rnode;
	struct pt_tuner *tuner = node.sysctl_data;
	struct pt_softc *sc = tuner->sc;
	u_int new;

	new = pt_tuner_error(sc, tuner);
	tuner->tuner_error += new - tuner->tuner_error_prev;
	tuner->tuner_error_prev = new;
	node.sysctl_data = &tuner->tuner_error;
	return sysctl_lookup(SYSCTLFN_CALL(&node));
}

void
pt_sysctl_init(sc)
	struct pt_softc *sc;
{
	const struct sysctlnode *node;
	int n = sc->sc_dev.dv_unit * 4;
	int ch, root;
	char name[16];

	sysctl_createv(NULL, 0, NULL, &node, 0, CTLTYPE_NODE, "pt", NULL,
	    NULL, 0, NULL, 0, CTL_HW, CTL_CREATE, CTL_EOL);
	root = node->sysctl_num;

	sysctl_createv(NULL, 0, NULL, &node, CTLFLAG_READWRITE,
	    CTLTYPE_INT, "lnb_mode", NULL, NULL, 0, &pt_lnb_mode, 0,
	    CTL_HW, root, CTL_CREATE, CTL_EOL);

#ifdef PT_STAT
	snprintf(name, sizeof name, "pages%d", sc->sc_dev.dv_unit);
	sysctl_createv(NULL, 0, NULL, &node, CTLFLAG_READWRITE,
	    CTLTYPE_INT, name, NULL, NULL, 0, &sc->sc_pages, 0,
	    CTL_HW, root, CTL_CREATE, CTL_EOL);

	for (ch = 0; ch < NTUNER; ch++) {
		snprintf(name, sizeof name, "fifo%d", n + ch);
		sysctl_createv(NULL, 0, NULL, &node, CTLFLAG_READWRITE,
		    CTLTYPE_INT, name, NULL, NULL, 0, &sc->sc_tuner[ch].used,
		    0, CTL_HW, root, CTL_CREATE, CTL_EOL);
	}
#endif

	for (ch = 0; ch < NTUNER; ch++) {
		snprintf(name, sizeof name, "error%d", n + ch);
		sysctl_createv(NULL, 0, NULL, &node, CTLFLAG_READWRITE,
		    CTLTYPE_INT, name, NULL, sysctl_pt_error, 0,
		    (void *)&sc->sc_tuner[ch], 0,
		    CTL_HW, root, CTL_CREATE, CTL_EOL);
	}
}

#define F_PT1 0x01
#define F_PT2 0x02

struct initdata {
	u_char reg;
	u_char val;
	u_char flags;
};

static inline int
init_isdb_s(sc, addr)
	struct pt_softc *sc;
	u_int addr;
{
	static struct initdata init_s[] = {
		{ 0x04, 0x02, F_PT1|F_PT2 },
		{ 0x0d, 0x55, F_PT1       },
		{ 0x11, 0x40, F_PT1|F_PT2 },
		{ 0x13, 0x80, F_PT1|F_PT2 },
		{ 0x17, 0x01, F_PT1|F_PT2 },
		{ 0x1c, 0x0a, F_PT1|F_PT2 },
		{ 0x1d, 0xaa, F_PT1|F_PT2 },
		{ 0x1e, 0x20, F_PT1|F_PT2 },
		{ 0x1f, 0x88, F_PT1|F_PT2 },
		{ 0x51, 0xb0, F_PT1|F_PT2 },
		{ 0x52, 0x89, F_PT1|F_PT2 },
		{ 0x53, 0xb3, F_PT1|F_PT2 },
		{ 0x5a, 0x2d, F_PT1|F_PT2 },
		{ 0x5b, 0xd3, F_PT1|F_PT2 },
		{ 0x85, 0x69, F_PT1|F_PT2 },
		{ 0x87, 0x04, F_PT1|F_PT2 },
		{ 0x8e, 0x26, F_PT1|F_PT2 },
		{ 0xa3, 0xf7, F_PT1|F_PT2 },
		{ 0xa5, 0xc0, F_PT1|F_PT2 },
		{ 0 }
	};
	struct initdata *p;
	struct i2c_cmd cmd;
	u_int afcg, flag = 0;

	cmd.len = 1;
	cmd.value[0] = 0x0f;	/* AFCG */
	afcg = i2c_read(sc, addr, &cmd, 1);

	switch (sc->sc_cardtype) {
	case CARD_TYPE_PT1:
		if (afcg != 0x4c) {	/* 20MHz */
			printf(": AFCG 0x%x\n", afcg);
			return -1;
		}
		flag = F_PT1;
		break;
	case CARD_TYPE_PT2:
		if (afcg != 0x52) {	/* 25MHz */
			printf(": AFCG 0x%x\n", afcg);
			return -1;
		}
		flag = F_PT2;
		break;
	}
	for (p = init_s; p->flags; p++) {
		if (p->flags & flag) {
			cmd.len = 2;
			cmd.value[0] = p->reg;
			cmd.value[1] = p->val;
			i2c_write(sc, addr, &cmd);
		}
	}
	return 0;
}

static inline int
init_isdb_t(sc, addr)
	struct pt_softc *sc;
	u_int addr;
{
	static struct initdata init_t[] = {
		{ 0x03, 0x90, F_PT1|F_PT2 },
		{ 0x14, 0x8f, F_PT1       },
		{ 0x1c, 0x2a, F_PT1|F_PT2 },
		{ 0x1d, 0xa8, F_PT1|F_PT2 },
		{ 0x1e, 0xa2, F_PT1|F_PT2 },
		{ 0x22, 0x83, F_PT1|F_PT2 },
		{ 0x31, 0x0d, F_PT1       },
		{ 0x32, 0xe0, F_PT1       },
		{ 0x39, 0xd3, F_PT1       },
		{ 0x3a, 0x00, F_PT1|F_PT2 },
		{ 0x5c, 0x40, F_PT1|F_PT2 },
		{ 0x5f, 0x80, F_PT1|F_PT2 },
		{ 0x75, 0x0a, F_PT1|F_PT2 },
		{ 0x76, 0x4c, F_PT1|F_PT2 },
		{ 0x77, 0x03, F_PT1|F_PT2 },
		{ 0xef, 0x01, F_PT1|F_PT2 },
		{ 0 }
	};
	struct initdata *p;
	struct i2c_cmd cmd;
	u_int flag = 0;

	switch (sc->sc_cardtype) {
	case CARD_TYPE_PT1:
		flag = F_PT1;
		break;
	case CARD_TYPE_PT2:
		flag = F_PT2;
		break;
	}
	for (p = init_t; p->flags; p++) {
		if (p->flags & flag) {
			cmd.len = 2;
			cmd.value[0] = p->reg;
			cmd.value[1] = p->val;
			i2c_write(sc, addr, &cmd);
		}
	}
	return 0;
}

int
pt_tuner_init(sc)
	struct pt_softc *sc;
{
	static struct i2c_cmd init_s = { 4, { 0xfe, 0xc0, 0xf0, 0x04 }};
	static struct i2c_cmd init_t =
	    { 7, { 0xfe, 0xc2, 0x01, 0x8f, 0xc1, 0x80, 0x80 }};
	struct i2c_cmd cmd;
	int i;
	u_int addr;

	i2c_init(sc);

	cmd.len = 2;
	cmd.value[0] = 0x01;
	cmd.value[1] = 0x80;
	for (i = 0; i < NTUNER; i++) {
		addr = sc->sc_tuner[i].addr;
		i2c_write(sc, addr, &cmd);
	}
	for (i = 0; i < NTUNER; i++) {
		addr = sc->sc_tuner[i].addr;
		switch (sc->sc_tuner[i].type) {
		case TUNER_TYPE_ISDB_S:
			if (init_isdb_s(sc, addr) < 0)
				return -1;
			i2c_write(sc, addr, &init_s);
			break;
		case TUNER_TYPE_ISDB_T:
			if (init_isdb_t(sc, addr) < 0)
				return -1;
			i2c_write(sc, addr, &init_t);
			break;
		}
	}
	if (sc->sc_cardtype == CARD_TYPE_PT1) {
		for (i = 0; i < NTUNER; i++) {
			switch (sc->sc_tuner[i].type) {
			case TUNER_TYPE_ISDB_S:
				s_tune(sc, &sc->sc_tuner[i], 0);
				break;
			case TUNER_TYPE_ISDB_T:
				t_tune(sc, &sc->sc_tuner[i], 0);
				break;
			}
		}
	}
	return 0;
}

inline void
s_sleep(sc, addr)
	struct pt_softc *sc;
	int addr;
{
	struct i2c_cmd cmd;

	cmd.len = 2;
	cmd.value[0] = 0x17;
	cmd.value[1] = 0x01;
	i2c_write(sc, addr, &cmd);
}

inline void
t_sleep(sc, addr)
	struct pt_softc *sc;
	int addr;
{
	struct i2c_cmd cmd;

	cmd.len = 2;
	cmd.value[0] = 0x03;
	cmd.value[1] = 0x90;
	i2c_write(sc, addr, &cmd);
}

inline void
s_wakeup(sc, addr)
	struct pt_softc *sc;
	int addr;
{
	struct i2c_cmd cmd;

	cmd.len = 2;
	cmd.value[0] = 0x17;
	cmd.value[1] = 0x00;
	i2c_write(sc, addr, &cmd);
}

inline void
t_wakeup(sc, addr)
	struct pt_softc *sc;
	int addr;
{
	struct i2c_cmd cmd;

	cmd.len = 2;
	cmd.value[0] = 0x03;
	cmd.value[1] = 0x80;
	i2c_write(sc, addr, &cmd);
}

void
pt_tuner_sleep(sc, tuner)
	struct pt_softc *sc;
	struct pt_tuner *tuner;
{
	switch (tuner->type) {
	case TUNER_TYPE_ISDB_S:
		s_sleep(sc, tuner->addr);
		return;
	case TUNER_TYPE_ISDB_T:
		t_sleep(sc, tuner->addr);
		return;
	}
}

void
pt_tuner_wakeup(sc, tuner)
	struct pt_softc *sc;
	struct pt_tuner *tuner;
{
	switch (tuner->type) {
	case TUNER_TYPE_ISDB_S:
		s_tune(sc, tuner, 0);
		s_wakeup(sc, tuner->addr);
		return;
	case TUNER_TYPE_ISDB_T:
		t_tune(sc, tuner, 0);
		t_wakeup(sc, tuner->addr);
		return;
	}
}

/*
 * Map channel number to frequency.
 * ch		ch		ch
 * 0  BS-1	12 ND-2		24 ND-1
 * 1  BS-3	13 ND-4		25 ND-3
 * 2  BS-5	14 ND-6		26 ND-5
 * ...		...		...
 * 11 BS-23	23 ND-24	35 ND-23
 */
inline u_int
s_frequency(channel)
	u_int channel;
{
	if (channel < 12) {
		/*
		 * BS-1 ... BS-23:
		 *   f = 1049.48 + 38.36 * channel;
		 */
		return (268667 + 9820 * channel + 128) / 256;
	}
	if (channel < 24) {
		/* ND-2, ND-4, ..., ND-24 */
		return 1593 + 20 * ((channel - 12) * 2 + 1);
	} else {
		/* ND-1, ND-3, ..., ND-23 */
		return 1593 + 20 * ((channel - 24) * 2);
	}
}

inline u_int
s_lookup_cb(f)
	u_int f;
{
	struct table {
		u_int freq:16;
		u_int value:8;
	};
	static struct table cb[] = {
		{  986, 0xb2 },
		{ 1072, 0xd2 },
		{ 1154, 0xe2 },
		{ 1291, 0x20 },
		{ 1447, 0x40 },
		{ 1615, 0x60 },
		{ 1791, 0x80 },
		{ 1972, 0xa0 },
		{   -1, 0xc0 }
	};
	struct table *t = cb;

	for (;;) {
		if (f < t->freq)
			return t->value;
		t++;
	}
}

int
s_tune(sc, tuner, wait)
	struct pt_softc *sc;
	struct pt_tuner *tuner;
	int wait;
{
	struct i2c_cmd cmd1, cmd2, cmd3;
	u_int addr = tuner->addr;
	u_int channel = tuner->current.freq;
	u_int f, c, val;
	int i;

	f = s_frequency(channel);
	c = s_lookup_cb(f);

	if (f < 1072)
		f = (f & ~0xf) << 1 | (f & 0xf);

	cmd1.len = 6;
	cmd1.value[0] = 0xfe;		/* THROUGH */
	cmd1.value[1] = 0xc0;		/* addr|0 (W) */
	cmd1.value[2] = 0x40 | f >> 8;
	cmd1.value[3] = f;
	cmd1.value[4] = 0xe0;
	cmd1.value[5] = c;

	cmd2.len = 3;
	cmd2.value[0] = 0xfe;
	cmd2.value[1] = 0xc0;
	cmd2.value[2] = 0xe4;

	cmd3.len = 4;
	cmd3.value[0] = 0xfe;
	cmd3.value[1] = 0xc0;
	cmd3.value[2] = 0xf4;
	cmd3.value[3] = c | 0x04;

	/* set frequency */
	i2c_write(sc, addr, &cmd1);
	i2c_write(sc, addr, &cmd2);
	i2c_write(sc, addr, &cmd3);

	/* Wait until the PLL is ready. */
	cmd1.len = 2;
	cmd1.value[0] = 0xfe;		/* THROUGH */
	cmd1.value[1] = 0xc1;		/* addr|1 (R) */
	for (i = 0; i < 100; i++) {
		val = i2c_read(sc, addr, &cmd1, 1);
		if (val & 0x40)
			goto lock_ok;
	}
	if (wait == 0)
		return 0;
	printf("%s: ISDB-S frequency error\n", sc->sc_dev.dv_xname);
	return -1;

lock_ok:
	cmd1.len = 2;
	cmd1.value[0] = 0x03;
	cmd1.value[1] = 0x01;
	i2c_write(sc, addr, &cmd1);
	if (wait == 0)
		return 0;

	cmd2.len = 1;
	cmd2.value[0] = 0xc3;
	for (i = 0; i < 20; i++) {
		tsleep(sc, 0, "s-tune", hz/8);
		val = i2c_read(sc, addr, &cmd2, 1);
		if ((val & 0x10) == 0)
			return 0;
	}
	printf("%s: ISDB-S failed to set channel %d\n",
	    sc->sc_dev.dv_xname, channel);
	return -1;
}

int
read_tsid(sc, tuner)
	struct pt_softc *sc;
	struct pt_tuner *tuner;
{
	struct i2c_cmd cmd;
	u_int slot = tuner->current.slot;
	u_int tsid;
	int i;

	if (slot > 7)
		return 0xffff;

	cmd.len = 1;
	cmd.value[0] = 0xce + slot * 2;
	for (i = 0; i < 10; i++) {
		tsid = i2c_read(sc, tuner->addr, &cmd, 2);
		if (tsid != 0)
			return tsid;
		tsleep(sc, 0, "rdtsid", hz/8);
	}
	printf("%s: failed to read TS ID\n", sc->sc_dev.dv_xname);
	return -1;
}

int
ts_lock(sc, tuner, ts_id)
	struct pt_softc *sc;
	struct pt_tuner *tuner;
	u_int ts_id;
{
	struct i2c_cmd cmd;
	u_int addr = tuner->addr;
	int i;

	/* set TS-ID */
	cmd.len = 3;
	cmd.value[0] = 0x8f;
	cmd.value[1] = ts_id >> 8;
	cmd.value[2] = ts_id;
	i2c_write(sc, addr, &cmd);

	/* check TS-ID */
	cmd.len = 1;
	cmd.value[0] = 0xe6;
	for (i = 0; i < 100; i++) {
		if (i2c_read(sc, addr, &cmd, 2) == ts_id)
			return 0;
	}
	printf("%s: failed to lock TS\n", sc->sc_dev.dv_xname);
	return -1;
}

static inline u_int
t_frequency(channel)
	u_int channel;
{
	u_int offset = 0;

	if (channel >= 12)
		offset += 2;
	if (channel >= 17)
		offset -= 2;
	if (channel >= 63)
		offset += 2;

	return 93 + channel * 6 + offset;
}

static inline u_int
t_lookup_cb(f)
	u_int f;
{
	struct table {
		u_int freq:16;
		u_int value:8;
	};
	static struct table cb[] = {
		{  90, 4 << 5 | 0 },
		{ 140, 4 << 5 | 1 },
		{ 170, 5 << 5 | 1 },
		{ 220, 3 << 5 | 2 },
		{ 330, 5 << 5 | 2 },
		{ 402, 7 << 5 | 2 },
		{ 450, 3 << 5 | 4 },
		{ 550, 4 << 5 | 4 },
		{ 600, 5 << 5 | 4 },
		{ 700, 6 << 5 | 4 },
		{  -1, 7 << 5 | 4 }
	};	/*	CP	BS */
	struct table *t = cb;

	for (;;) {
		if (f < t->freq)
			return t->value;
		t++;
	}
}

int
t_tune(sc, tuner, wait)
	struct pt_softc *sc;
	struct pt_tuner *tuner;
	int wait;
{
	struct i2c_cmd cmd;
	u_int addr = tuner->addr;
	u_int channel = tuner->current.freq;
	int freq_offset = tuner->current.slot;
	u_int f, c, val;
	int i;

	f = t_frequency(channel);
	c = t_lookup_cb(f);
	val = f * 7 + 400 + freq_offset;

	cmd.len = 6;
	cmd.value[0] = 0xfe;		/* THROUGH */
	cmd.value[1] = 0xc2;		/* addr|0 (W) */
	cmd.value[2] = val >> 8;	/* DB1 */
	cmd.value[3] = val;		/* DB2 */
	cmd.value[4] = 0x80;		/* CB1 */
	cmd.value[5] = c;		/* CB2 */
	i2c_write(sc, addr, &cmd);

	/* Wait until the PLL is ready. */
	cmd.len = 2;
	cmd.value[0] = 0xfe;		/* THROUGH */
	cmd.value[1] = 0xc3;		/* addr|1 (R) */
	for (i = 0; i < 100; i++) {
		val = i2c_read(sc, addr, &cmd, 1);
		if (val == 0xff) {
			printf("pt: value 0xff\n");
			continue;
		}
		if (val & 0x40)		/* FL */
			goto lock_ok;
	}
	if (wait == 0)
		return 0;
	printf("%s: ISDB-T frequency error\n", sc->sc_dev.dv_xname);
	return -1;

lock_ok:
	cmd.len = 2;
	cmd.value[0] = 0x01;
	cmd.value[1] = 0x40;
	i2c_write(sc, addr, &cmd);
	if (wait == 0)
		return 0;

	cmd.len = 1;
	cmd.value[0] = 0x80;
	for (i = 0; i < 20; i++) {
		tsleep(sc, 0, "t-tune", hz/8);
		val = i2c_read(sc, addr, &cmd, 1);
		if ((val & 0x08) == 0)
			return 0;
	}
	printf("%s: ISDB-T failed to set channel %d\n",
	    sc->sc_dev.dv_xname, channel);
	return -1;
}

int
s_cnr(sc, addr)
	struct pt_softc *sc;
	u_int addr;
{
	struct i2c_cmd cmd;

	cmd.len = 1;
	cmd.value[0] = 0xbc;
	return i2c_read(sc, addr, &cmd, 2);
}

int
t_cnr(sc, addr)
	struct pt_softc *sc;
	u_int addr;
{
	struct i2c_cmd cmd;

	cmd.len = 1;
	cmd.value[0] = 0x8b;
	return i2c_read(sc, addr, &cmd, 3);
}

int
s_agc(sc, addr)
	struct pt_softc *sc;
	u_int addr;
{
	struct i2c_cmd cmd;

	cmd.len = 1;
	cmd.value[0] = 0xba;
	return i2c_read(sc, addr, &cmd, 1);
}

int
t_agc(sc, addr)
	struct pt_softc *sc;
	u_int addr;
{
	struct i2c_cmd cmd;

	/* ifagc_dt */
	cmd.len = 1;
	cmd.value[0] = 0x82;
	return i2c_read(sc, addr, &cmd, 1);
}

int
s_carrier(sc, addr)
	struct pt_softc *sc;
	u_int addr;
{
	struct i2c_cmd cmd;

	/* afcfrq */
	cmd.len = 1;
	cmd.value[0] = 0xbb;
	return (int8_t)i2c_read(sc, addr, &cmd, 1);
}

int
t_carrier(sc, addr)
	struct pt_softc *sc;
	int addr;
{
	struct i2c_cmd cmd;

	/* carafc_dt */
	cmd.len = 1;
	cmd.value[0] = 0x84;
	return (int16_t)i2c_read(sc, addr, &cmd, 2);
}

int
s_clock(sc, addr)
	struct pt_softc *sc;
	u_int addr;
{
	struct i2c_cmd cmd;

	/* clkfrq */
	cmd.len = 1;
	cmd.value[0] = 0xbe;
	return (int8_t)i2c_read(sc, addr, &cmd, 1);
}

int
t_clock(sc, addr)
	struct pt_softc *sc;
	u_int addr;
{
	struct i2c_cmd cmd;

	/* clkafc_dt */
	cmd.len = 1;
	cmd.value[0] = 0x86;
	return (int16_t)i2c_read(sc, addr, &cmd, 2);
}

int
s_raw_err(sc, addr)
	struct pt_softc *sc;
	u_int addr;
{
	struct i2c_cmd cmd;

	cmd.len = 1;
	cmd.value[0] = 0xf5;
	return i2c_read(sc, addr, &cmd, 3);
}

int
t_raw_err(sc, addr)
	struct pt_softc *sc;
	u_int addr;
{
	struct i2c_cmd cmd;

	cmd.len = 1;
	cmd.value[0] = 0x97;
	return i2c_read(sc, addr, &cmd, 3);
}

#define I2C_WRITE 0
#define I2C_READ  1
#define SCL_H 1
#define SCL_L 0
#define SDA_H 1
#define SDA_L 0

#define state(sc, scl, sda) do { \
	pt_i2c(sc, (sc)->sc_ramaddr, (sc)->sc_ramaddr + 1, 0, scl, sda, 1); \
	(sc)->sc_ramaddr++; \
} while (0)
#define state_read(sc) do { \
	pt_i2c(sc, (sc)->sc_ramaddr, (sc)->sc_ramaddr + 1, 1, SCL_H, SDA_H, 1); \
	(sc)->sc_ramaddr++; \
} while (0)
#define state_free(sc) do { \
	pt_i2c(sc, (sc)->sc_ramaddr, 0, 0, SCL_H, SDA_H, 0); \
	(sc)->sc_ramaddr++; \
} while (0)
#define state_end(sc) do { \
	pt_i2c(sc, (sc)->sc_ramaddr, 0, 0, SCL_H, SDA_H, 1); \
} while (0)

void inline
pt_i2c(sc, addr, next, rw, scl, sda, busy)
	struct pt_softc *sc;
	u_int addr, next, rw, scl, sda, busy;
{
	u_int val;

	sda ^= 1;
	scl ^= 1;
	val = addr << 18 | busy << 13 | rw << 12 | scl << 11 | sda << 10 | next;
	PT_WRITE(sc, PT_CODE_REG, val);
	/*
	NEXT	0x000003ff
	SDA	0x00000400
	SCL	0x00000800
	RW	0x00001000
	BUSY	0x00002000
	ADDR	0x0ffc0000
	*/
}

static inline u_int64_t
xc3s_read_id(sc)
	struct pt_softc *sc;
{
	int i;
	u_int64_t id = 0;

	for (i = 0; i < 57; i++) {
		if (PT_READ(sc, PT_STATUS_REG) & ID_SYNC)
			break;
		PT_WRITE(sc, PT_COMMAND_REG, ID_PULSE);
	}

	for (i = 0; i < 57; i++) {
		if (PT_READ(sc, PT_STATUS_REG) & ID_DATA)
			id |= 1LL << i;
		PT_WRITE(sc, PT_COMMAND_REG, ID_PULSE);
	}
	return id;
}

static inline int
xc3s_unlock(sc)
	struct pt_softc *sc;
{
	int i;

	PT_WRITE(sc, PT_COMMAND_REG, ID_PULSE);

	for (i = 0; i < 3; i++) {
		if (PT_READ(sc, PT_STATUS_REG) & UNLOCKED)
			return 0;
		delay(10);
	}
	return -1;
}

static inline int
xc3s_pci_reset(sc)
	struct pt_softc *sc;
{
	int i;

	PT_WRITE(sc, PT_COMMAND_REG, ON(PCI_RESET));
	PT_WRITE(sc, PT_COMMAND_REG, OFF(PCI_RESET));
	for (i = 0; i < 100; i++) {
		if (PT_READ(sc, PT_STATUS_REG) & PCI_LOCKED)
			return 0;
		delay(10);
	}
	return -1;
}

static inline int
xc3s_ram_reset(sc)
	struct pt_softc *sc;
{
	int i;

	PT_WRITE(sc, PT_COMMAND_REG, ON(RAM_RESET));
	PT_WRITE(sc, PT_COMMAND_REG, OFF(RAM_RESET));
	for (i = 0; i < 100; i++) {
		if (PT_READ(sc, PT_STATUS_REG) & RAM_LOCKED)
			return 0;
		delay(10);
	}
	return -1;
}

static inline int
xc3s_ram_shift(sc, dir)
	struct pt_softc *sc;
	u_int dir;
{
	u_int done;
	int i;

	done = PT_READ(sc, PT_STATUS_REG) & RAM_SHIFT_DONE;

	PT_WRITE(sc, PT_COMMAND_REG, dir | RAM_ENABLE);
	for (i = 0; i < 1024; i++) {
		if ((PT_READ(sc, PT_STATUS_REG) & RAM_SHIFT_DONE) != done)
			return 0;
	}
	return -1;
}

int
xc3s_init(sc)
	struct pt_softc *sc;
{
	int i, phase;

	/* Initialize I2C RAM. */
	for (i = 0; i < 1024; i++)
		pt_i2c(sc, i, 0, 0, SCL_H, SDA_H, 0);
	/* Wait 512 PCI clocks. */
	for (i = 0; i < 512 / 4; i++)
		pt_i2c(sc, 0, 0, 0, SCL_H, SDA_H, 0);

	sc->sc_id = xc3s_read_id(sc);

	if (xc3s_unlock(sc)) {
		printf(": failed to unlock xc3s\n");
		return -1;
	}
	if (xc3s_pci_reset(sc)) {
		printf(": failed to reset PCI\n");
		return -1;
	}
	if (xc3s_ram_reset(sc)) {
		printf(": failed to reset RAM\n");
		return -1;
	}

	switch(sc->sc_cardtype) {
	default:
	case CARD_TYPE_PT1:
		phase = 128;
		break;
	case CARD_TYPE_PT2:
		phase = 166;
		break;
	}
	while (phase-- > 0) {
		if (xc3s_ram_shift(sc, RAM_SHIFT_DEC)) {
			printf(": xc3s_ram_shift timed out\n");
			return -1;
		}
	}

	/* Disable stream. */
	PT_WRITE(sc, PT_TEST_EN_REG, OFF(0xf));

	return 0;
}

/*
 * SCL = L ǸƤФơ1ӥå
 * SCL = L 
 */
static void
writebit(sc, val)
	struct pt_softc *sc;
	u_int val;
{
	state(sc, SCL_L, val);		/* SDA: new value */
	state(sc, SCL_H, val);		/* SCL: L -> H */
	state(sc, SCL_L, val);		/* SCL: H -> L */
}

/*
 * SCL = L ǸƤФơ1ӥåɤ
 * SCL = L 
 */
static void
readbit(sc)
	struct pt_softc *sc;
{
	state(sc, SCL_L, SDA_H);		/* SDA: H */
	state(sc, SCL_H, SDA_H);		/* SCL: L -> H */
	state_read(sc);
	state(sc, SCL_L, SDA_H);		/* SCL: H -> L */
}

void
sendbyte(sc, val)
	struct pt_softc *sc;
	u_int val;
{
	int i;

	for (i = 0; i < 8; i++) {
		writebit(sc, (val & 0x80) ? 1 : 0);
		val <<= 1;
	}
}

static void
i2c_init(sc)
	struct pt_softc *sc;
{
	pt_i2c(sc, 0, 0, 0, SCL_H, SDA_H, 0);	/* free */
	pt_i2c(sc, 1, 2, 0, SCL_H, SDA_H, 1);	/* busy */
	pt_i2c(sc, 2, 3, 0, SCL_H, SDA_L, 1);	/* SDA: H -> L */
	pt_i2c(sc, 3, 4, 0, SCL_H, SDA_H, 1);	/* SDA: L -> H */
	pt_i2c(sc, 4, 0, 0, SCL_H, SDA_H, 1);	/* end */

	PT_WRITE(sc, PT_COMMAND_REG, I2C_EXEC);
	delay(50);
}

static void
i2c_begin(sc)
	struct pt_softc *sc;
{
	sc->sc_ramaddr = 0;
	state_free(sc);				/* free */
	state(sc, SCL_H, SDA_H);		/* busy */
}

/*
 * Make a START condition.  Assume:
 *   a. SCL = SDA = H (repeat = 0)  or
 *   b. SCL = L       (repeat = 1)
 */
static void
i2c_start(sc, repeat)
	struct pt_softc *sc;
	int repeat;
{
	if (repeat) {
		state(sc, SCL_L, SDA_H);	/* SDA: H */
		state(sc, SCL_H, SDA_H);	/* SCL: L -> H */
	}
	state(sc, SCL_H, SDA_L);		/* SDA: H -> L */
	state(sc, SCL_L, SDA_L);		/* SCL: H -> L */
}

/*
 * Make a STOP condition.  Assume called with SCL = L.
 */
static void
i2c_stop(sc)
	struct pt_softc *sc;
{
	state(sc, SCL_L, SDA_L);		/* SDA: L */
	state(sc, SCL_H, SDA_L);		/* SCL: L -> H */
	state_end(sc);				/* SDA: L -> H */
}

static inline void
i2c_prepare_write(sc, addr, cmd)
	struct pt_softc *sc;
	u_int addr;
	struct i2c_cmd *cmd;
{
	int i;

	i2c_begin(sc);
	i2c_start(sc, 0);

	/* Write the slave address and R/W# . */
	sendbyte(sc, addr | I2C_WRITE);

	/* Slave ACK (always H). */
	writebit(sc, 1);

	/* Send data. */
	for (i = 0; i < cmd->len; i++) {
		sendbyte(sc, cmd->value[i]);

		/* Slave ACK (always H). */
		writebit(sc, 1);
	}

	i2c_stop(sc);
}

static inline void
i2c_prepare_read(sc, addr, cmd, rlen)
	struct pt_softc *sc;
	u_int addr;
	struct i2c_cmd *cmd;
	int rlen;
{
	int i;

	i2c_begin(sc);
	i2c_start(sc, 0);

	/* Write the slave address and R/W#. */
	sendbyte(sc, addr | I2C_WRITE);

	/* Slave ACK (always H). */
	writebit(sc, 1);

	/* Send data. */
	for (i = 0; i < cmd->len; i++) {
		sendbyte(sc, cmd->value[i]);

		/* Slave ACK (always H). */
		writebit(sc, 1);
	}

	/* Repeated START. */
	i2c_start(sc, 1);

	/* Write the slave address and R/W#. */
	sendbyte(sc, addr | I2C_READ);

	/* Slave ACK (always H). */
	writebit(sc, 1);

	/* Receive data. */
	for (;;) {
		for (i = 0; i < 8; i++)
			readbit(sc);
		rlen--;
		if (rlen <= 0) {
			/* Send NACK */
			writebit(sc, 1);
			break;
		} else {
			/* Send ACK */
			writebit(sc, 0);
		}
	}

	i2c_stop(sc);
}

void
i2c_write(sc, addr, cmd)
	struct pt_softc *sc;
	u_int addr;
	struct i2c_cmd *cmd;
{
	int i;

#ifdef DIAGNOSTIC
	if (cmd->len > sizeof cmd->value)
		panic("i2c_write");
#endif
	mutex_enter(&sc->sc_lock);

	i2c_prepare_write(sc, addr, cmd);
	PT_WRITE(sc, PT_COMMAND_REG, I2C_EXEC);

	for (i = 0; i < 200; i++) {
		delay(10);
		if ((PT_READ(sc, PT_STATUS_REG) & I2C_BUSY) == 0)
			goto done;
	}
	printf("%s: i2c_write: timeout\n", sc->sc_dev.dv_xname);
done:
	mutex_exit(&sc->sc_lock);
}

u_int
i2c_read(sc, addr, cmd, size)
	struct pt_softc *sc;
	u_int addr;
	struct i2c_cmd *cmd;
	int size;
{
	int i;
	u_int val;

#ifdef DIAGNOSTIC
	if (cmd->len > sizeof cmd->value)
		panic("i2c_read");
#endif
	mutex_enter(&sc->sc_lock);

	i2c_prepare_read(sc, addr, cmd, size);
	PT_WRITE(sc, PT_COMMAND_REG, I2C_EXEC);

	for (i = 0; i < 200; i++) {
		delay(10);
		if ((PT_READ(sc, PT_STATUS_REG) & I2C_BUSY) == 0)
			goto done;
	}
	printf("%s: i2c_read: timeout\n", sc->sc_dev.dv_xname);
done:
	val = PT_READ(sc, PT_I2CREAD_REG);
	mutex_exit(&sc->sc_lock);
	val &= 0xffffffff >> (32 - size * 8);
	return val;
}
