/*	$Id: pt3.c,v 1.7 2015/09/16 03:11:56 tsubai Exp $	*/

/*-
 * Copyright (c) 2013-2015 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/callout.h>
#include <sys/conf.h>
#include <sys/device.h>
#include <sys/ioctl.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>

#include <machine/bus.h>

#include <dev/pci/pcivar.h>
#include <dev/pci/pcireg.h>
#include <dev/pci/pcidevs.h>

#include "pt3reg.h"
#include "ptio.h"

#if defined(__powerpc__) || defined(__powerpc64__)
#define PT_SAVE_KVA	/* Save KVA space. */
#endif

#define NBUFS 32
#define PT3_BLOCK_SIZE (188 * 1024)	/* multiple of 188 */
#define PT3_STAT

struct pt3_tuner {
	struct pt3_softc *softc;
	u_char fe_addr;		/* TC90522 frontend I2C address */
	u_char tuner_addr;	/* QM1D1C0042 or MxL301RF I2C address */
	u_int tuner_error;
	u_int xfer_stat;

	paddr_t dma_desc_pa;
	paddr_t dma_buffer_pa;
	u_char (*dma_buffer)[PT3_BLOCK_SIZE];
	u_int dma_pos;
	u_int dma_ch;
	u_int opened:1;
	u_int type:1;
	u_int unit:1;
	u_int fast_search:1;	/* ISDB-S */
	u_int freq;		/* ISDB-T */
};

struct pt3_softc {
	struct device sc_dev;
	bus_space_tag_t sc_regt;
	bus_space_handle_t sc_regh;
	bus_space_tag_t sc_memt;
	bus_space_handle_t sc_memh;
	bus_dma_tag_t sc_dmat;

	struct kmutex sc_lock;
	struct kmutex sc_i2c_lock;

	struct pt3_tuner sc_tuner[4];

	u_int sc_i2c_cmdlen;
	u_int sc_i2c_lastcmd;

	int sc_lnbcnt;
	struct callout sc_lnb_ch;
	u_int sc_bus_count;
};

int pt3_match(struct device *, struct cfdata *, void *);
void pt3_attach(struct device *, struct device *, void *);

static void pt3_enable_lnb(struct pt3_softc *, struct pt3_tuner *);
static void pt3_disable_lnb(struct pt3_softc *, struct pt3_tuner *);
static void lnb_off(void *);
static int pt3_dma_setup(struct pt3_softc *, struct pt3_tuner *);
static void pt3_dma_start(struct pt3_softc *, struct pt3_tuner *);
static void pt3_dma_stop(struct pt3_softc *, struct pt3_tuner *);
static int pt3_dma_ready(struct pt3_tuner *);
static int pt3_tuner_sleep(struct pt3_softc *, struct pt3_tuner *);
static int pt3_tuner_wakeup(struct pt3_softc *, struct pt3_tuner *);
static int getcnr(struct pt3_tuner *);
static int getagc(struct pt3_tuner *);
static int getcarrier(struct pt3_tuner *);
static int getclock(struct pt3_tuner *);
static u_int tuner_error(struct pt3_softc *, struct pt3_tuner *);
static void *dma_alloc(struct pt3_softc *, long, paddr_t *);
static void pt3_sysctl_init(struct pt3_softc *);

static int pt3_tuner_init(struct pt3_softc *);
static int s_set_channel(struct pt3_softc *, struct pt3_tuner *,
    struct pt_channel *);
static int t_lock_wait(struct pt3_tuner *);
static int t_set_channel(struct pt3_softc *, struct pt3_tuner *,
    struct pt_channel *);

static void i2c_cmd(struct pt3_softc *, u_int);
static void i2c_start(struct pt3_softc *);
static void i2c_stop(struct pt3_softc *);
static void i2c_write(struct pt3_softc *, u_char *, int);
static void i2c_read(struct pt3_softc *, int);
static void i2c_init(struct pt3_softc *);
static void i2c_send(struct pt3_softc *, u_int, void *, int);
static void i2c_recv(struct pt3_softc *, u_int, int);
static void i2c_result(struct pt3_softc *, u_char *, int);
static int i2c_run(struct pt3_softc *);
static int i2c_kick(struct pt3_softc *, u_int);
static int i2c_reset(struct pt3_softc *);

static void delay_ms(u_int);
static int fe_write(struct pt3_tuner *, int, void *, int);
static int fe_read(struct pt3_tuner *, int, void *, int);
static int fe_write1(struct pt3_tuner *, u_int, u_int);
static int pskmsrst(struct pt3_tuner *);
static int imsrst(struct pt3_tuner *);
static int fe_s_init(struct pt3_tuner *);
static int fe_t_init(struct pt3_tuner *);
static int s_pins_normal(struct pt3_tuner *);
static int t_pins_normal(struct pt3_tuner *);
static int fe_set_power(struct pt3_tuner *, int, int);
static int s_enable_agc(struct pt3_tuner *);
static int s_disable_agc(struct pt3_tuner *);
static int t_enable_agc(struct pt3_tuner *);
static int t_disable_agc(struct pt3_tuner *);
static int set_ts(struct pt3_tuner *, u_int);
static int read_tsid(struct pt3_tuner *);
static int read_retryov_tmunvld_fulock(struct pt3_tuner *, int *, int *, int *);
static int s_read_agc(struct pt3_tuner *);
static int t_read_ifagc(struct pt3_tuner *);
static int t_read_carafc(struct pt3_tuner *);
static int t_read_clkafc(struct pt3_tuner *);
static int s_read_cn(struct pt3_tuner *);
static int t_read_cndat(struct pt3_tuner *);

static int qm_write(struct pt3_tuner *, u_int, u_int);
static int qm_write_n(struct pt3_tuner *, u_int, u_char *, int);
static int qm_read(struct pt3_tuner *, u_int);
static int qm_sleep(struct pt3_tuner *);
static int qm_wakeup(struct pt3_tuner *);
static int qm_fast_search_mode(struct pt3_tuner *, int);
static int s_init(struct pt3_tuner *);
static int s_sleep(struct pt3_tuner *);
static int s_wakeup(struct pt3_tuner *);
static void qm_pll_param(u_int, u_int *, u_int *, u_int *);
static int qm_tune(struct pt3_tuner *, u_int, int);
static u_int s_channel_freq(u_int);
static int s_lock_status(struct pt3_tuner *, int *);
static int s_tune(struct pt3_tuner *, u_int);

static int mxl_write(struct pt3_tuner *, void *, int);
static int mxl_read(struct pt3_tuner *, u_int, void *);
static int mxl_setreg(struct pt3_tuner *, u_int, u_int);
static int mxl_rftune(struct pt3_tuner *, u_int);
static u_int t_channel_freq(u_int);
static int t_lock_status(struct pt3_tuner *, int *, int *);
static int t_tune(struct pt3_tuner *, u_int, int);
static int slptim(struct pt3_tuner *, int);
static int t_sleep(struct pt3_tuner *);
static int t_wakeup(struct pt3_tuner *);

#define PT_READ(sc, reg) bus_space_read_4((sc)->sc_regt, (sc)->sc_regh, reg)
#define PT_WRITE(sc, reg, val) \
    bus_space_write_4((sc)->sc_regt, (sc)->sc_regh, reg, val)

CFATTACH_DECL(pt, sizeof(struct pt3_softc), pt3_match, pt3_attach, NULL);

extern struct cfdriver pt_cd;

dev_type_open(pt3open);
dev_type_close(pt3close);
dev_type_read(pt3read);
dev_type_ioctl(pt3ioctl);
dev_type_mmap(pt3mmap);

const struct cdevsw pt_cdevsw = {
	pt3open, pt3close, pt3read, nullwrite, pt3ioctl,
	nostop, notty, nopoll, pt3mmap, nokqfilter
};

int pt3_lnb_mode = LNB_15V;

enum { TUNER_TYPE_ISDB_S, TUNER_TYPE_ISDB_T };

static struct {
	u_char type;
	u_char unit;
	u_char fe_addr;
	u_char tuner_addr;
	u_char dma_channel;
} tuners[] = {
	{ TUNER_TYPE_ISDB_S, 0, 0x22, 0xc6, 0 },
	{ TUNER_TYPE_ISDB_T, 0, 0x20, 0xc4, 2 },
	{ TUNER_TYPE_ISDB_S, 1, 0x26, 0xc0, 1 },
	{ TUNER_TYPE_ISDB_T, 1, 0x24, 0xc2, 3 }
};

#define DEVNAME(sc) ((sc)->sc_dev.dv_xname)

int
pt3_match(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{
	struct pci_attach_args *pa = aux;
	pci_chipset_tag_t pc = pa->pa_pc;
	pcitag_t tag = pa->pa_tag;
	pcireg_t subsys;

	if (PCI_VENDOR(pa->pa_id) == PCI_VENDOR_ALTERA &&
	    PCI_PRODUCT(pa->pa_id) == PCI_PRODUCT_ALTERA_EP4CGX15BF14C8N) {
		subsys = pci_conf_read(pc, tag, PCI_SUBSYS_ID_REG);
		if (PCI_SUBSYS_VENDOR(subsys) == 0xee8d)
			return 1;
	}
	return 0;
}

void
pt3_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct pt3_softc *sc = device_private(self);
	struct pci_attach_args *pa = aux;
	u_int csr, class;
	u_int x, ptn, fpga, maxbits;
	int i;
	struct pt3_tuner *tuner;

	csr = pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_COMMAND_STATUS_REG);
	csr |= PCI_COMMAND_MASTER_ENABLE;
	pci_conf_write(pa->pa_pc, pa->pa_tag, PCI_COMMAND_STATUS_REG, csr);

	class = pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_CLASS_REG);
	if (PCI_REVISION(class) != 1) {
		printf(": unknown revision\n");
		return;
	}

	if (pci_mapreg_map(pa, 0x10,
	    PCI_MAPREG_TYPE_MEM | PCI_MAPREG_MEM_TYPE_64BIT, 0,
	    &sc->sc_regt, &sc->sc_regh, NULL, NULL)) {
		printf(": failed to map registers\n");
		return;
	}
	if (pci_mapreg_map(pa, 0x18,
	    PCI_MAPREG_TYPE_MEM | PCI_MAPREG_MEM_TYPE_64BIT, 0,
	    &sc->sc_memt, &sc->sc_memh, NULL, NULL)) {
		printf(": failed to map i2c memory\n");
		return;
	}
	sc->sc_dmat = pa->pa_dmat;

	x = PT_READ(sc, PT3_VERSION);
	ptn = VERSION_PT(x);
	fpga = VERSION_FPGA(x);
	if (ptn != 3) {
		printf(": unsupported PT (%d)\n", ptn);
		return;
	}
	printf(": PT%d", ptn);
	if (fpga != 4) {
		printf(", with unsupported FPGA version 0x%x\n", fpga);
		return;
	}

	x = PT_READ(sc, PT3_SYSTEM);
	if ((x & SYSTEM_DS28E01) == 0) {
		printf(", not authorized\n");
		return;
	}
	maxbits = x & SYSTEM_MAXBITS;
	if ((1 << maxbits) - 8 < MAXDMASIZE) {
		printf(", DMA size is too small\n");
		return;
		/* maxbits = 20, i.e. 1MB */
	}
	printf("\n");

	mutex_init(&sc->sc_lock, MUTEX_DEFAULT, IPL_NONE);
	mutex_init(&sc->sc_i2c_lock, MUTEX_DEFAULT, IPL_NONE);
	callout_init(&sc->sc_lnb_ch, 0);

	for (i = 0; i < 4; i++) {
		tuner = &sc->sc_tuner[i];
		tuner->type = tuners[i].type;
		tuner->unit = tuners[i].unit;
		tuner->softc = sc;
		tuner->fe_addr = tuners[i].fe_addr;
		tuner->tuner_addr = tuners[i].tuner_addr;
		tuner->dma_ch = tuners[i].dma_channel;
		if (pt3_dma_setup(sc, tuner))
			return;
	}
	printf("%s: %d KB buffer\n", DEVNAME(sc),
	    PT3_BLOCK_SIZE / 1024 * NBUFS * 4);

	/* Make sure LNB power is off. */
	PT_WRITE(sc, PT3_LNBCTRL, LNBCTRL_EN_VALID | LNBCTRL_15V_VALID);

	pt3_sysctl_init(sc);

	if (pt3_tuner_init(sc)) {
		for (i = 0; i < 4; i++) {
			tuner = &sc->sc_tuner[i];
			pt3_dma_stop(sc, tuner);
			pt3_tuner_sleep(sc, tuner);
		}
		fe_set_power(&sc->sc_tuner[3], 0, 0);
		return;
	}

	/* Clear error counters. */
	for (i = 0; i < 4; i++) {
		tuner = &sc->sc_tuner[i];
		PT_WRITE(sc, PT3_TSCTRL(tuner->dma_ch), TSCTRL_RSTERR);
	}
}

void
pt3_enable_lnb(sc, tuner)
	struct pt3_softc *sc;
	struct pt3_tuner *tuner;
{
	u_int val;

	if (tuner->type != TUNER_TYPE_ISDB_S)
		return;

	mutex_enter(&sc->sc_lock);
	if (sc->sc_lnbcnt == 0) {
		callout_stop(&sc->sc_lnb_ch);
		switch (pt3_lnb_mode) {
		case LNB_11V:
			val = LNBCTRL_EN;
			break;
		case LNB_15V:
			val = LNBCTRL_EN | LNBCTRL_15V;
			break;
		default:
			val = 0;
		}
		val |= LNBCTRL_15V_VALID | LNBCTRL_EN_VALID;
		PT_WRITE(sc, PT3_LNBCTRL, val);
		/* printf("%s: LNB on\n", DEVNAME(sc)); */

		tsleep(sc, 0, "lnbwait", hz/2);
	}
	sc->sc_lnbcnt++;
	mutex_exit(&sc->sc_lock);
}

void
pt3_disable_lnb(sc, tuner)
	struct pt3_softc *sc;
	struct pt3_tuner *tuner;
{
	if (tuner->type != TUNER_TYPE_ISDB_S)
		return;

	mutex_enter(&sc->sc_lock);
	sc->sc_lnbcnt--;
	if (sc->sc_lnbcnt == 0)
		callout_reset(&sc->sc_lnb_ch, hz * 40, lnb_off, sc);
	mutex_exit(&sc->sc_lock);
}

void
lnb_off(v)
	void *v;
{
	struct pt3_softc *sc = v;

	PT_WRITE(sc, PT3_LNBCTRL, LNBCTRL_EN_VALID | LNBCTRL_15V_VALID);
	/* printf("%s: LNB off\n", DEVNAME(sc)); */
}

int
pt3_dma_setup(sc, tuner)
	struct pt3_softc *sc;
	struct pt3_tuner *tuner;
{
	long buffer_size = PT3_BLOCK_SIZE * NBUFS;
	struct dmadesc *p;
	paddr_t addr, next;
	int ndesc, len;

	tuner->dma_buffer = dma_alloc(sc, buffer_size, &tuner->dma_buffer_pa);
	if (tuner->dma_buffer == NULL) {
		printf("%s: failed to allocate DMA buffer\n", DEVNAME(sc));
		return -1;
	}

	ndesc = (buffer_size + MAXDMASIZE - 1) / MAXDMASIZE;
	p = dma_alloc(sc, ndesc * sizeof *p, &tuner->dma_desc_pa);
	if (p == NULL) {
		printf("%s: failed to allocate DMA descriptors\n", DEVNAME(sc));
		return -1;
	}

	/* Create DMA descriptors. */
	addr = tuner->dma_buffer_pa;
	next = tuner->dma_desc_pa + sizeof *p;
	while (buffer_size > 0) {
		len = MIN(MAXDMASIZE, buffer_size);
		buffer_size -= len;
		if (buffer_size == 0)
			next = tuner->dma_desc_pa;
#ifdef __LP64__
		p->buffer_l = htole32(addr);
		p->buffer_h = htole32(addr >> 32);
		p->len      = htole32(len);
		p->next_l   = htole32(next);
		p->next_h   = htole32(next >> 32);
#else
		p->buffer_l = htole32(addr);
		p->buffer_h = 0;
		p->len      = htole32(len);
		p->next_l   = htole32(next);
		p->next_h   = 0;
#endif
		addr += len;
		next += sizeof *p;
		p++;
	}
	return 0;
}

void
pt3_dma_start(sc, tuner)
	struct pt3_softc *sc;
	struct pt3_tuner *tuner;
{
	paddr_t start_addr = tuner->dma_desc_pa;
	u_int ch = tuner->dma_ch;
	int i;

	/* Mark all buffers free. */
	for (i = 0; i < NBUFS; i++)
		tuner->dma_buffer[i][0] = 0;
	tuner->dma_pos = 0;

#ifdef __LP64__
	PT_WRITE(sc, PT3_DMADESC_H(ch), start_addr >> 32);
#else
	PT_WRITE(sc, PT3_DMADESC_H(ch), 0);
#endif
	PT_WRITE(sc, PT3_DMADESC_L(ch), start_addr);
	PT_WRITE(sc, PT3_DMACTRL(ch), DMACTRL_START);
}

void
pt3_dma_stop(sc, tuner)
	struct pt3_softc *sc;
	struct pt3_tuner *tuner;
{
	u_int ch = tuner->dma_ch;
	int i;

	PT_WRITE(sc, PT3_DMACTRL(ch), DMACTRL_STOP);
	for (i = 0; i < 1000; i++) {
		if ((PT_READ(sc, PT3_STATUS(ch)) & STATUS_DMA) == 0)
			break;
		delay(1);
	}
}

int
pt3_dma_ready(tuner)
	struct pt3_tuner *tuner;
{
	struct pt3_softc *sc;
	u_int next, status;
	u_char *p;

	next = (tuner->dma_pos + 1) % NBUFS;
	p = tuner->dma_buffer[next];
	if (*p == 0x47)
		return 1;
	if (*p == 0)
		return 0;

	sc = tuner->softc;
	status = PT_READ(sc, PT3_STATUS(tuner->dma_ch));

	/* Reset DMA. */
	pt3_dma_stop(sc, tuner);
	pt3_dma_start(sc, tuner);

	printf("%s: corrupted DMA data (tuner %d:%d)\n",
	    DEVNAME(sc), tuner->type, tuner->unit);

	printf("status 0x%08x\n", status);
	if (status & 0x000040)
		printf("internal FIFO A overflow\n");
	if (status & 0x000080)
		printf("internal FIFO A underflow\n");
	if (status & 0x040000)
		printf("internal FIFO B overflow\n");
	if (status & 0x080000)
		printf("internal FIFO B underflow\n");
	return 0;
}

int
pt3_tuner_sleep(sc, tuner)
	struct pt3_softc *sc;
	struct pt3_tuner *tuner;
{
	int error;

	if (tuner->type == TUNER_TYPE_ISDB_S)
		error = s_sleep(tuner);
	else
		error = t_sleep(tuner);
	/* delay(50000); */
	return error;
}

int
pt3_tuner_wakeup(sc, tuner)
	struct pt3_softc *sc;
	struct pt3_tuner *tuner;
{
	int error;

	if (tuner->type == TUNER_TYPE_ISDB_S)
		error = s_wakeup(tuner);
	else
		error = t_wakeup(tuner);
	return error;
}

static inline int
getcnr(tuner)
	struct pt3_tuner *tuner;
{
	if (tuner->type == TUNER_TYPE_ISDB_S)
		return s_read_cn(tuner);
	else
		return t_read_cndat(tuner);
}

static inline int
getagc(tuner)
	struct pt3_tuner *tuner;
{
	if (tuner->type == TUNER_TYPE_ISDB_S)
		return s_read_agc(tuner);
	else
		return t_read_ifagc(tuner);
}

static inline int
getcarrier(tuner)
	struct pt3_tuner *tuner;
{
	if (tuner->type == TUNER_TYPE_ISDB_S)
		return 0;
	else
		return t_read_carafc(tuner);
}

static inline int
getclock(tuner)
	struct pt3_tuner *tuner;
{
	if (tuner->type == TUNER_TYPE_ISDB_S)
		return 0;
	else
		return t_read_clkafc(tuner);
	return 0;
}

static u_int
tuner_error(sc, tuner)
	struct pt3_softc *sc;
	struct pt3_tuner *tuner;
{
	u_int gray, val;
	int i;

	gray = PT_READ(sc, PT3_TSERROR(tuner->dma_ch));
	for (val = 0, i = 31; i >= 0; i--)
		val |= (gray ^ (val >> 1)) & (1 << i);
	return val;
}

#if 0
/*
 * ä1.5ܤ餤®ɡ
 */
static u_int
tuner_error(sc, tuner)
	struct pt3_softc *sc;
	struct pt3_tuner *tuner;
{
	static u_char table[] = {
		0x0, 0x1, 0x3, 0x2, 0x7, 0x6, 0x4, 0x5,
		0xf, 0xe, 0xc, 0xd, 0x8, 0x9, 0xb, 0xa
	};
	u_int val = 0, rev = 0;
	u_int n;
	int i;

	gray = PT_READ(sc, PT3_TSERROR(tuner->dma_ch));
	for (i = 28; i >= 0; i -= 4) {
		n = table[(gray >> i) & 0xf];
		val |= (n ^ rev) << i;
		if (n & 1)
			rev ^= 0xf;
	}
	return val;
}
#endif

#define PTUNIT(x) (minor(x) >> 2)
#define PTCHAN(x) (minor(x) & 3)

int
pt3open(dev, flag, mode, l)
	dev_t dev;
	int flag, mode;
	struct lwp *l;
{
	struct pt3_softc *sc = device_lookup_private(&pt_cd, PTUNIT(dev));
	struct pt3_tuner *tuner;

	if (sc == NULL)
		return ENXIO;
	tuner = &sc->sc_tuner[PTCHAN(dev)];
	if (tuner->opened)
		return EBUSY;

	tuner->opened = 1;
	pt3_enable_lnb(sc, tuner);
	pt3_tuner_wakeup(sc, tuner);
	tsleep(sc, 0, "ptopen", hz/8);
	return 0;
}

int
pt3close(dev, flag, mode, l)
	dev_t dev;
	int flag, mode;
	struct lwp *l;
{
	struct pt3_softc *sc = device_lookup_private(&pt_cd, PTUNIT(dev));
	struct pt3_tuner *tuner = &sc->sc_tuner[PTCHAN(dev)];

	tuner->tuner_error += tuner_error(sc, tuner);
	pt3_dma_stop(sc, tuner);

	pt3_tuner_sleep(sc, tuner);
	tsleep(sc, 0, "ptclose", hz/2);
	pt3_disable_lnb(sc, tuner);
	PT_WRITE(sc, PT3_TSCTRL(tuner->dma_ch), TSCTRL_RSTERR);
	tuner->opened = 0;
	return 0;
}

int
pt3read(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	struct pt3_softc *sc = device_lookup_private(&pt_cd, PTUNIT(dev));
	struct pt3_tuner *tuner = &sc->sc_tuner[PTCHAN(dev)];
	int error, i;
	u_int prev;
	u_char *p;
#ifdef PT3_STAT
	u_int n, opos = tuner->dma_pos;
#endif

	if (uio->uio_resid < PT3_BLOCK_SIZE)
		return EINVAL;

	/* XXX Should not error - i.e. wait forever? */
	for (i = 0; i < 20; i++) {
		if (pt3_dma_ready(tuner))
			goto ready;
		tsleep(tuner, 0, "ptwait", hz / 16);
	}
	return EIO;
ready:
	prev = (tuner->dma_pos - 1) % NBUFS;
	if (tuner->dma_buffer[prev][0] != 0)
		printf("%s: DMA overrun detected\n", DEVNAME(sc));

	while (uio->uio_resid >= PT3_BLOCK_SIZE) {
		p = tuner->dma_buffer[tuner->dma_pos];
		error = uiomove(p, PT3_BLOCK_SIZE, uio);
		if (error)
			return error;
		p[0] = 0;
		tuner->dma_pos = (tuner->dma_pos + 1) % NBUFS;
		if (!pt3_dma_ready(tuner))
			break;
	}
#ifdef PT3_STAT
	n = (tuner->dma_pos - opos) % NBUFS;
	if (n > tuner->xfer_stat)
		tuner->xfer_stat = n;
#endif
	return 0;
}

int
pt3ioctl(dev, cmd, data, flag, l)
	dev_t dev;
	u_long cmd;
	void *data;
	int flag;
	struct lwp *l;
{
	struct pt3_softc *sc = device_lookup_private(&pt_cd, PTUNIT(dev));
	struct pt3_tuner *tuner = &sc->sc_tuner[PTCHAN(dev)];
	int error, val, i;

	switch (cmd) {
	case PTIOSCHAN:
	case SET_CHANNEL:
		if (tuner->type == TUNER_TYPE_ISDB_S)
			error = s_set_channel(sc, tuner, data);
		else
			error = t_set_channel(sc, tuner, data);
		if (error)
			return error;

		/* Wait for the tuner stable. */
		for (i = 0; i < 10; i++) {
			PT_WRITE(sc, PT3_TSCTRL(tuner->dma_ch), TSCTRL_RSTERR);
			tsleep(tuner, 0, "pt3ioctl", hz/16);
			if (tuner_error(sc, tuner) == 0)
				break;
		}
		return 0;
	case PTIORECORD:
		if (*(int *)data)
			goto start_dma;
		else
			goto stop_dma;
	case START_REC:
	start_dma:
		pt3_dma_start(sc, tuner);
		return 0;
	case STOP_REC:
	stop_dma:
		pt3_dma_stop(sc, tuner);
		return 0;
	case PTIOGCNR:
	case GET_SIGNAL_STRENGTH:
		val = getcnr(tuner);
		if (val < 0)
			return EIO;
		*(int *)data = val;
		return 0;
	case PTIOGAGC:
		val = getagc(tuner);
		if (val < 0)
			return EIO;
		*(int *)data = val;
		return 0;
	case PTIOGCAR:
		val = getcarrier(tuner);
		if (val < 0)
			return EIO;
		*(int *)data = (int16_t)val;
		return 0;
	case PTIOGCLK:
		val = getclock(tuner);
		if (val < 0)
			return EIO;
		*(int *)data = (int16_t)val;
		return 0;
	case LNB_ENABLE:
	case LNB_DISABLE:
		return 0;
	}
	return EINVAL;
}

paddr_t
pt3mmap(dev, off, prot)
	dev_t dev;
	off_t off;
	int prot;
{
	struct pt3_softc *sc = device_lookup_private(&pt_cd, PTUNIT(dev));

	if (off < 0)
		return -1;
	if (off >= (off_t)PT3_BLOCK_SIZE * NBUFS)
		return -1;
	return sc->sc_tuner[PTCHAN(dev)].dma_buffer_pa + off;
}

void *
dma_alloc(sc, size, pap)
	struct pt3_softc *sc;
	long size;
	paddr_t *pap;
{
	bus_dma_tag_t t = sc->sc_dmat;
	bus_dmamap_t map;
	bus_dma_segment_t segs[1];
	void *kva;
	int n;
#ifdef __LP64__
	bus_size_t b = 0x100000000;	/* Don't cross 4GB boundary */
#else
	bus_size_t b = 0;
#endif

	/* Allocate and map a contiguous memory region. */
	if (bus_dmamem_alloc(t, size, NBPG, b, segs, 1, &n, BUS_DMA_NOWAIT))
		goto out1;
#ifdef PT_SAVE_KVA
	*pap = segs[0].ds_addr;
	kva = (void *)segs[0].ds_addr;
	bzero(kva, size);
	return kva;
#endif
	if (bus_dmamem_map(t, segs, 1, size, &kva, BUS_DMA_NOWAIT))
		goto out2;
	if (bus_dmamap_create(t, size, 1, size, 0, BUS_DMA_NOWAIT, &map))
		goto out3;
	if (bus_dmamap_load(t, map, kva, size, NULL, BUS_DMA_NOWAIT))
		goto out4;

	bzero(kva, size);
	*pap = map->dm_segs[0].ds_addr;
	bus_dmamap_unload(t, map);
	bus_dmamap_destroy(t, map);
	return kva;

out4:
	bus_dmamap_destroy(t, map);
out3:
	bus_dmamem_unmap(t, kva, size);
out2:
	bus_dmamem_free(t, segs, 1);
out1:
	return NULL;
}

static int
sysctl_pt_error(SYSCTLFN_ARGS)
{
	struct sysctlnode node = *rnode;
	struct pt3_tuner *tuner = node.sysctl_data;
	struct pt3_softc *sc = tuner->softc;

	tuner->tuner_error += tuner_error(sc, tuner);
	PT_WRITE(sc, PT3_TSCTRL(tuner->dma_ch), TSCTRL_RSTERR);
	node.sysctl_data = &tuner->tuner_error;
	return sysctl_lookup(SYSCTLFN_CALL(&node));
}

static int
sysctl_pt_bus(SYSCTLFN_ARGS)
{
	struct sysctlnode node = *rnode;
	struct pt3_softc *sc = node.sysctl_data;

	sc->sc_bus_count = PT_READ(sc, PT3_BUS);
	node.sysctl_data = &sc->sc_bus_count;
	return sysctl_lookup(SYSCTLFN_CALL(&node));
}

void
pt3_sysctl_init(sc)
	struct pt3_softc *sc;
{
	const struct sysctlnode *node;
	int n = sc->sc_dev.dv_unit * 4;
	int ch, root;
	char name[16];

	sysctl_createv(NULL, 0, NULL, &node, 0, CTLTYPE_NODE, "pt3", NULL,
	    NULL, 0, NULL, 0, CTL_HW, CTL_CREATE, CTL_EOL);
	root = node->sysctl_num;

	sysctl_createv(NULL, 0, NULL, &node, CTLFLAG_READWRITE,
	    CTLTYPE_INT, "lnb_mode", NULL, NULL, 0, &pt3_lnb_mode, 0,
	    CTL_HW, root, CTL_CREATE, CTL_EOL);

#ifdef PT3_STAT
	for (ch = 0; ch < 4; ch++) {
		snprintf(name, sizeof name, "max%d", n + ch);
		sysctl_createv(NULL, 0, NULL, &node, CTLFLAG_READWRITE,
		    CTLTYPE_INT, name, NULL, NULL, 0, &sc->sc_tuner[ch].xfer_stat,
		    0, CTL_HW, root, CTL_CREATE, CTL_EOL);
	}
#endif

	for (ch = 0; ch < 4; ch++) {
		snprintf(name, sizeof name, "error%d", n + ch);
		sysctl_createv(NULL, 0, NULL, &node, CTLFLAG_READWRITE,
		    CTLTYPE_INT, name, NULL, sysctl_pt_error, 0,
		    (void *)&sc->sc_tuner[ch], 0,
		    CTL_HW, root, CTL_CREATE, CTL_EOL);
	}
	snprintf(name, sizeof name, "bus%d", sc->sc_dev.dv_unit);
	sysctl_createv(NULL, 0, NULL, &node, 0,
	    CTLTYPE_INT, name, NULL, sysctl_pt_bus, 0,
	    (void *)sc, 0,
	    CTL_HW, root, CTL_CREATE, CTL_EOL);
}

static inline int
pskmsrst(tuner)
	struct pt3_tuner *tuner;
{
	return fe_write1(tuner, 0x03, 0x01);
}

static inline int
imsrst(tuner)
	struct pt3_tuner *tuner;
{
	return fe_write1(tuner, 0x01, 0x40);
}

static inline int
fe_s_init(tuner)
	struct pt3_tuner *tuner;
{
	if (pskmsrst(tuner))
		return -1;
	return fe_write1(tuner, 0x1e, 0x10);
}

static inline int
fe_t_init(tuner)
	struct pt3_tuner *tuner;
{
	if (imsrst(tuner))
		return -1;
	return fe_write1(tuner, 0x1c, 0x10);
}

static inline int
s_pins_normal(tuner)
	struct pt3_tuner *tuner;
{
	if (fe_write1(tuner, 0x1c, 0x15) ||
	    fe_write1(tuner, 0x1f, 0x04))
		return -1;
	return 0;
}

static inline int
t_pins_normal(tuner)
	struct pt3_tuner *tuner;
{
	return fe_write1(tuner, 0x1d, 0x01);
}

int
pt3_tuner_init(sc)
	struct pt3_softc *sc;
{
	struct pt3_tuner *tuner;
	int i, channel, error;

	if (i2c_reset(sc))
		return -1;

	for (i = 0; i < 4; i++) {
		tuner = &sc->sc_tuner[i];
		if (tuner->type == TUNER_TYPE_ISDB_S)
			error = fe_s_init(tuner);
		else
			error = fe_t_init(tuner);
		if (error)
			printf("%s: failed to initialize frontend\n", DEVNAME(sc));
	}

	if (fe_set_power(&sc->sc_tuner[3], 1, 0)) {
		printf("%s: failed to enable tuner power\n", DEVNAME(sc));
		return -1;
	}

	for (i = 0; i < 4; i++) {
		tuner = &sc->sc_tuner[i];
		if (tuner->type == TUNER_TYPE_ISDB_S)
			error = s_pins_normal(tuner);
		else
			error = t_pins_normal(tuner);
		if (error)
			printf("%s: failed to set pins\n", DEVNAME(sc));
	}
	delay(1000);

	for (i = 0; i < 4; i++) {
		tuner = &sc->sc_tuner[i];
		if (tuner->type == TUNER_TYPE_ISDB_S) {
			if (s_init(tuner))
				printf("%s: failed to initialize tuner\n", DEVNAME(sc));
		}
	}

#define INIT_ADDR0 (4096 + 0)
#define INIT_ADDR1 (4096 + 2042)

	/* Execute initialization sequence from the internal ROM. */
	if (i2c_kick(sc, INIT_ADDR1)) {
		printf("%s: internal init failed\n", DEVNAME(sc));
		return -1;
	}
	if (fe_set_power(&sc->sc_tuner[3], 1, 1)) {
		printf("%s: failed to enable amp power\n", DEVNAME(sc));
		return -1;
	}

	for (i = 0; i < 4; i++) {
		tuner = &sc->sc_tuner[i];
		if (pt3_tuner_wakeup(sc, tuner)) {
			printf("%s: failed to wakeup tuner\n", DEVNAME(sc));
			return -1;
		}
		delay(50000);
		if (tuner->type == TUNER_TYPE_ISDB_S) {
			error = s_tune(tuner, 0);
		} else {
			channel = (tuner->unit == 0) ? 70 : 71;
			error = t_tune(tuner, channel, 0);
		}
		if (error)
			printf("%s: failed to set frequency\n", DEVNAME(sc));
		if (pt3_tuner_sleep(sc, tuner)) {
			printf("%s: tuner didn't sleep\n", DEVNAME(sc));
			return -1;
		}
	}
	return 0;
}

static int
s_set_channel(sc, tuner, ch)
	struct pt3_softc *sc;
	struct pt3_tuner *tuner;
	struct pt_channel *ch;
{
	u_int channel = ch->freq;
	u_int slot = ch->slot;
	int i, tsid, val;

	if (channel > 35)
		return EINVAL;
	if (slot > 7)
		return EINVAL;
	if (s_tune(tuner, channel))
		return EIO;

	for (i = 0; i < 20; i++) {
		tsleep(tuner, 0, "s-tune1", hz / 16);
		tsid = set_ts(tuner, slot);
		if (tsid != -1)
			goto got_tsid;
	}
	printf("%s: failed to set channel ISDB-S %d:%d\n", DEVNAME(sc),
	    channel, slot);
	return EIO;
got_tsid:
	for (i = 0; i < 20; i++) {
		val = read_tsid(tuner);
		if (val < 0)
			return EIO;
		if ((val & 0xffff) == tsid)
			return 0;
		tsleep(tuner, 0, "s-tune2", hz / 16);
	}
	return EIO;
}

static inline int
t_lock_wait(tuner)
	struct pt3_tuner *tuner;
{
	int retryov, tmunvld, fulock;
	int i;

	for (i = 0; i < 30; i++) {
		if (read_retryov_tmunvld_fulock(tuner, &retryov,
		    &tmunvld, &fulock))
			return -1;
		if (fulock == 0)
			return 0;
		if (retryov)
			return -1;
		tsleep(tuner, 0, "t-tune1", hz / 16);
	}
	return -1;
}

static int
t_set_channel(sc, tuner, ch)
	struct pt3_softc *sc;
	struct pt3_tuner *tuner;
	struct pt_channel *ch;
{
	u_int channel = ch->freq;
	int offset = ch->slot;
	int i;

	if (channel > 112)
		return EINVAL;
	if (abs(offset) > 28)
		return EINVAL;
	for (i = 0; i < 3; i++) {
		if (t_tune(tuner, channel, offset))
			continue;
		if (t_lock_wait(tuner) == 0)
			return 0;
	}
	printf("%s: failed to set channel ISDB-T %d\n", DEVNAME(sc), channel);
	return EIO;
}

#define PT_READ_RAM(sc, off) \
    bus_space_read_1((sc)->sc_memt, (sc)->sc_memh, 2048 + (off))
#define PT_WRITE_RAM(sc, off, val) \
    bus_space_write_1((sc)->sc_memt, (sc)->sc_memh, 2048 + (off), val)

#define OP_END        0
#define OP_ADDRESS    1
#define OP_SCL_L      2
#define OP_SCL_H      3
#define OP_SDA_L      4
#define OP_SDA_H      5
#define OP_RESET      6
#define OP_SLEEP      7
#define OP_SDA_L_NOP  8
#define OP_SDA_L_READ 9		/* not for normal use */
#define OP_SDA_L_ACK0 10	/* not for normal use */
#define OP_SDA_L_ACK1 11	/* not for normal use */
#define OP_SDA_H_NOP  12
#define OP_SDA_H_READ 13
#define OP_SDA_H_ACK0 14
#define OP_SDA_H_ACK1 15

static void
i2c_cmd(sc, opcode)
	struct pt3_softc *sc;
	u_int opcode;
{
	if (sc->sc_i2c_cmdlen >= 4096)
		panic("%s: I2C cmd too long", DEVNAME(sc));
	if (sc->sc_i2c_cmdlen % 2 == 0)
		sc->sc_i2c_lastcmd = opcode;
	else {
		opcode <<= 4;
		opcode |= sc->sc_i2c_lastcmd;
		PT_WRITE_RAM(sc, sc->sc_i2c_cmdlen / 2, opcode);
	}
	sc->sc_i2c_cmdlen++;
}

static void
i2c_start(sc)
	struct pt3_softc *sc;
{
	i2c_cmd(sc, OP_SDA_H);
	i2c_cmd(sc, OP_SCL_H);
	i2c_cmd(sc, OP_SDA_L);
	i2c_cmd(sc, OP_SCL_L);
}

static void
i2c_stop(sc)
	struct pt3_softc *sc;
{
	i2c_cmd(sc, OP_SDA_L);
	i2c_cmd(sc, OP_SCL_H);
	i2c_cmd(sc, OP_SDA_H);
}

static void
i2c_write(sc, data, len)
	struct pt3_softc *sc;
	u_char *data;
	int len;
{
	int i;

	for (; len > 0; len--, data++) {
		for (i = 0; i < 8; i++) {
			if (*data & (0x80 >> i))
				i2c_cmd(sc, OP_SDA_H_NOP);
			else
				i2c_cmd(sc, OP_SDA_L_NOP);
		}
		i2c_cmd(sc, OP_SDA_H_ACK0);
	}
}

static void
i2c_read(sc, len)
	struct pt3_softc *sc;
	int len;
{
	int i;

	for (; len > 0; len--) {
		for (i = 0; i < 8; i++)
			i2c_cmd(sc, OP_SDA_H_READ);
		if (len == 1)
			i2c_cmd(sc, OP_SDA_H_NOP);	/* NAK */
		else
			i2c_cmd(sc, OP_SDA_L_NOP);	/* ACK */
	}
}

static void inline
i2c_init(sc)
	struct pt3_softc *sc;
{
	sc->sc_i2c_cmdlen = 0;
}

static void
i2c_send(sc, addr, buf, len)
	struct pt3_softc *sc;
	u_int addr;
	void *buf;
	int len;
{
	u_char a[1];

	a[0] = addr;
	i2c_start(sc);
	i2c_write(sc, a, 1);
	i2c_write(sc, buf, len);
}

static void
i2c_recv(sc, addr, len)
	struct pt3_softc *sc;
	u_int addr;
	int len;
{
	u_char a[1];

	a[0] = addr | 1;
	i2c_start(sc);
	i2c_write(sc, a, 1);
	i2c_read(sc, len);
}

static void
i2c_result(sc, p, len)
	struct pt3_softc *sc;
	u_char *p;
	int len;
{
	u_int addr = 0;

	for (; len > 0; len--, addr++)
		*p++ = PT_READ_RAM(sc, addr);
}

static int
i2c_run(sc)
	struct pt3_softc *sc;
{
	i2c_stop(sc);
	i2c_cmd(sc, OP_END);
	i2c_cmd(sc, OP_END);
	return i2c_kick(sc, 0);
}

static int
i2c_kick(sc, addr)
	struct pt3_softc *sc;
	u_int addr;
{
	u_int x;
	int i;

	if (PT_READ(sc, PT3_I2C_STATUS) & I2C_STATUS_RUN) {
		printf("%s: I2C busy\n", DEVNAME(sc));
		return -1;
	}

	PT_WRITE(sc, PT3_I2C_CMD, I2C_CMD_EXEC | addr);
	for (i = 0; i < 2000; i++) {
		x = PT_READ(sc, PT3_I2C_STATUS);
		if ((x & I2C_STATUS_RUN) == 0)
			goto done;
		delay(10);
	}
	printf("%s: I2C timeout\n", DEVNAME(sc));
	return -1;
done:
	if (x & (I2C_STATUS_ACK0 | I2C_STATUS_ACK1)) {
#if 0
		if (x & I2C_STATUS_ACK0)
			printf("i2c_kick: ACK0\n");
		if (x & I2C_STATUS_ACK1)
			printf("i2c_kick: ACK1\n");
#endif
		return -1;
	}
	return 0;
}

static int
i2c_reset(sc)
	struct pt3_softc *sc;
{
	int i;

	PT_WRITE(sc, PT3_I2C_CMD, I2C_CMD_RESET);
	for (i = 0; i < 100; i++) {
		if ((PT_READ(sc, PT3_I2C_STATUS) & I2C_STATUS_RUN) == 0)
			return 0;
		delay(1000);
	}
	printf("%s: failed to reset I2C\n", DEVNAME(sc));
	return -1;
}

static inline void
delay_ms(ms)
	u_int ms;
{
	delay(ms * 1000);
}

#define THROUGH 0xfe

static int
fe_write(tuner, reg, data, len)
	struct pt3_tuner *tuner;
	int reg, len;
	void *data;
{
	struct pt3_softc *sc = tuner->softc;
	int error;
	u_char buf[16];

	mutex_enter(&sc->sc_i2c_lock);
	i2c_init(sc);

	buf[0] = reg;
	bcopy(data, buf + 1, len);
	i2c_send(sc, tuner->fe_addr, buf, len + 1);
	error = i2c_run(sc);
	mutex_exit(&sc->sc_i2c_lock);
	return error;
}

static int
fe_read(tuner, reg, data, len)
	struct pt3_tuner *tuner;
	int reg, len;
	void *data;
{
	struct pt3_softc *sc = tuner->softc;
	int error;
	u_char buf[1];

	mutex_enter(&sc->sc_i2c_lock);
	i2c_init(sc);

	buf[0] = reg;
	i2c_send(sc, tuner->fe_addr, buf, 1);
	i2c_recv(sc, tuner->fe_addr, len);
	error = i2c_run(sc);
	i2c_result(sc, data, len);
	mutex_exit(&sc->sc_i2c_lock);
	return error;
}

static int
fe_write1(tuner, reg, val)
	struct pt3_tuner *tuner;
	u_int reg, val;
{
	u_char buf[1];

	buf[0] = val;
	return fe_write(tuner, reg, buf, 1);
}

static int
fe_set_power(tuner, tuner_power, amp_power)
	struct pt3_tuner *tuner;
	int tuner_power, amp_power;
{
	u_char val = 0x11;

	val |= tuner_power ? 0xc0 : 0x80;
	val |= amp_power ? 0x0c : 0x08;
	return fe_write1(tuner, 0x1e, val);
}

static int
s_enable_agc(tuner)
	struct pt3_tuner *tuner;
{
	u_char val1[1], val2[2];

	val1[0] = 0xff;
	val2[0] = 0x01 | (tuner->unit == 0 ? 0xb0 : 0x30);
	val2[1] = 0x40;
	if (fe_write(tuner, 0x0a, val1, 1) ||
	    fe_write(tuner, 0x10, val2, 2))
		return -1;
	return pskmsrst(tuner);
}

static int
s_disable_agc(tuner)
	struct pt3_tuner *tuner;
{
	u_char val1[1], val2[2];

	val1[0] = 0x00;
	val2[0] = tuner->unit == 0 ? 0xb0 : 0x30;
	val2[1] = 0x00;
	if (fe_write(tuner, 0x0a, val1, 1) ||
	    fe_write(tuner, 0x10, val2, 2))
		return -1;
	return pskmsrst(tuner);
}

static int
t_enable_agc(tuner)
	struct pt3_tuner *tuner;
{
	if (fe_write1(tuner, 0x25, 0x40) ||
	    fe_write1(tuner, 0x23, 0x4c))
		return -1;
	return imsrst(tuner);
}

static int
t_disable_agc(tuner)
	struct pt3_tuner *tuner;
{
	if (fe_write1(tuner, 0x25, 0x00) ||
	    fe_write1(tuner, 0x23, 0x4d))
		return -1;
	return imsrst(tuner);
}

static int
set_ts(tuner, slot)
	struct pt3_tuner *tuner;
	u_int slot;
{
	int tsid;
	u_char val[2];

	if (fe_read(tuner, 0xc3, val, 1))
		return -1;
	if (val[0] & 0x10)
		return -1;

	/* Read TS-ID. */
	if (fe_read(tuner, 0xce + slot * 2, val, 2))
		return -1;
	tsid = val[0] << 8 | val[1];
	if (tsid == 0)
		return -1;

	/* Set it. */
	if (fe_write(tuner, 0x8f, val, 2))
		return -1;

	return tsid;
}

static int
read_tsid(tuner)
	struct pt3_tuner *tuner;
{
	u_char val[2];

	if (fe_read(tuner, 0xe6, val, 2))
		return -1;
	return val[0] << 8 | val[1];
}

static int
read_retryov_tmunvld_fulock(tuner, retryov, tmunvld, fulock)
	struct pt3_tuner *tuner;
	int *retryov, *tmunvld, *fulock;
{
	u_char val;

	if (fe_read(tuner, 0x80, &val, 1))
		return -1;
	*retryov = (val & 0x80) ? 1 : 0;
	*tmunvld = (val & 0x20) ? 1 : 0;
	*fulock  = (val & 0x08) ? 1 : 0;
	return 0;
}

static int
s_read_agc(tuner)
	struct pt3_tuner *tuner;
{
	u_char val;

	if (fe_read(tuner, 0xba, &val, 1))
		return -1;
	return val & 0x7f;
}

static int
t_read_ifagc(tuner)
	struct pt3_tuner *tuner;
{
	u_char val;

	if (fe_read(tuner, 0x82, &val, 1))
		return -1;
	return val;
}

static int
t_read_carafc(tuner)
	struct pt3_tuner *tuner;
{
	u_char val[2];

	if (fe_read(tuner, 0x84, &val, 2))
		return -1;
	return val[0] << 8 | val[1];
}

static int
t_read_clkafc(tuner)
	struct pt3_tuner *tuner;
{
	u_char val[2];

	if (fe_read(tuner, 0x86, &val, 2))
		return -1;
	return val[0] << 8 | val[1];
}

static int
s_read_cn(tuner)
	struct pt3_tuner *tuner;
{
	u_char val[2];

	if (fe_read(tuner, 0xbc, val, 2))
		return -1;
	return val[0] << 8 | val[1];
}

static int
t_read_cndat(tuner)
	struct pt3_tuner *tuner;
{
	u_char val[3];

	if (fe_read(tuner, 0x8b, val, 3))
		return -1;
	return val[0] << 16 | val[1] << 8 | val[2];
}

#if 0
#define tc90522_master_clock_freq 78	/* MHz */
#endif

#define qm_crystal_freq 16000	/* kHz */
#define lpf_wait 20
#define fast_search_wait 4
#define normal_search_wait 15

#define INIT_DUMMY_RESET 0x0c

static int
qm_write(tuner, reg, data)
	struct pt3_tuner *tuner;
	u_int reg, data;
{
	u_char val = data;

	return qm_write_n(tuner, reg, &val, 1);
}

static int
qm_write_n(tuner, reg, data, len)
	struct pt3_tuner *tuner;
	u_int reg;
	u_char *data;
	int len;
{
	struct pt3_softc *sc = tuner->softc;
	int error;
	u_char buf[32];

	if (len > 29)
		panic("qm_write");

	mutex_enter(&sc->sc_i2c_lock);
	i2c_init(sc);

	buf[0] = THROUGH;
	buf[1] = tuner->tuner_addr;
	buf[2] = reg;
	bcopy(data, buf + 3, len);
	i2c_send(sc, tuner->fe_addr, buf, len + 3);
	error = i2c_run(sc);
	mutex_exit(&sc->sc_i2c_lock);
	return error;
}

static int
qm_read(tuner, reg)
	struct pt3_tuner *tuner;
	u_int reg;
{
	struct pt3_softc *sc = tuner->softc;
	int error;
	u_char buf[3];

	mutex_enter(&sc->sc_i2c_lock);
	i2c_init(sc);

	buf[0] = THROUGH;
	buf[1] = tuner->tuner_addr;
	buf[2] = reg;
	i2c_send(sc, tuner->fe_addr, buf, 3);
	buf[0] = THROUGH;
	buf[1] = tuner->tuner_addr | 1;
	i2c_send(sc, tuner->fe_addr, buf, 2);
	i2c_recv(sc, tuner->fe_addr, 1);
	error = i2c_run(sc);
	i2c_result(sc, buf, 1);
	mutex_exit(&sc->sc_i2c_lock);
	if (error)
		return -1;
	return buf[0];
}

static inline int
qm_sleep(tuner)
	struct pt3_tuner *tuner;
{
	if (qm_write(tuner, 0x05, 0xcd) ||
	    qm_write(tuner, 0x01, 0x15))
		return -1;
	return 0;
}

static inline int
qm_wakeup(tuner)
	struct pt3_tuner *tuner;
{
	if (qm_write(tuner, 0x01, 0x1c) ||
	    qm_write(tuner, 0x05, 0xc5))
		return -1;
	return 0;
}

static inline int
qm_fast_search_mode(tuner, mode)
	struct pt3_tuner *tuner;
	int mode;
{
	u_int val;

	if (mode) {
		tuner->fast_search = 1;
		val = 0x11;
	} else {
		tuner->fast_search = 0;
		val = 0x10;
	}
	if (qm_write(tuner, 0x03, val))
		return -1;
	return 0;
}

static u_char init_reg[0x20] = {
	0x48, 0x1c, 0xa0, 0x10, 0xbc, 0xc5, 0x20, 0x33,
	0x06, 0x00, 0x00, 0x00, 0x43, 0x00, 0x00, 0x00,
	0x00, 0xff, 0xf3, 0x00, 0x2a, 0x64, 0xa6, 0x86,
	0x8c, 0xcf, 0xb8, 0xf1, 0xa8, 0xf2, 0x89, 0x00
};

static int
s_init(tuner)
	struct pt3_tuner *tuner;
{
	qm_write(tuner, 0x01, INIT_DUMMY_RESET);	/* ignore error */
	qm_write(tuner, 0x01, INIT_DUMMY_RESET);	/* ignore error */

	/* soft reset */
	if (qm_write(tuner, 0x01, INIT_DUMMY_RESET))
		return -1;
	delay(1000);
	if (qm_write(tuner, 0x01, 0x1c))
		return -1;

	/* Check ID. */
	if (qm_read(tuner, 0x00) != 0x48)
		return -1;

	/* LPF tuning on */
	delay_ms(1);
	if (qm_write(tuner, 0x0c, 0x43))
		return -1;
	delay_ms(lpf_wait);

	if (qm_write_n(tuner, 0x01, &init_reg[0x01], 12) ||
	    qm_write_n(tuner, 0x11, &init_reg[0x11], 15))
		return -1;

	if (qm_wakeup(tuner))
		return -1;
	if (qm_fast_search_mode(tuner, 0))
		return -1;

	return 0;
}

static int
s_sleep(tuner)
	struct pt3_tuner *tuner;
{
	if (s_disable_agc(tuner))
		return -1;
	qm_sleep(tuner);
	fe_write1(tuner, 0x17, 1);
	return 0;
}

static int
s_wakeup(tuner)
	struct pt3_tuner *tuner;
{
	fe_write1(tuner, 0x17, 0);
	qm_wakeup(tuner);
	return 0;
}

#define ROUND_DIV(a, b) (((a) + ((a) >= 0 ? (b) : -(b)) / 2) / (b))
/* #define ROUND_DIV2(a, b) (((a) - ((a) >= 0 ? 0 : (b) - 1)) / (b)) */

static inline void
qm_pll_param(freq, sd, N, A)
	u_int freq;
	u_int *sd, *N, *A;
{
	u_int a;
	int b;

	/*
	 * original code looks like:
	 *
	 *	double M = (double)freq / qm_crystal_freq;
	 *	u_int a = M + 0.5;			# round off
	 *	double b = M - a;
	 *
	 *	u_int N = (a - 12) / 4;
	 *	u_int A = a - 4 * (N + 1) - 5;
	 *	u_int sd = (2^20 * b) & 0x3fffff;	# signed 2.20 format
	 */
	a = ROUND_DIV(freq, qm_crystal_freq);
	b = freq - a * qm_crystal_freq;
	*sd = ROUND_DIV((1LL << 20) * b, qm_crystal_freq);
	*sd &= 0x3fffff;		/* signed 2.20 format */

	*N = (a - 12) / 4;
	*A = a - 4 * (*N + 1) - 5;
}

static struct {
	u_int f:24;
	u_int val:8;
} vco_table[9] = {
	{ 2151000, 0xf0 },
	{ 1950000, 0xe0 },
	{ 1800000, 0xd0 },
	{ 1600000, 0xc0 },
	{ 1450000, 0xb0 },
	{ 1250000, 0xa0 },
	{ 1200000, 0x70 },
	{  975000, 0x60 },
	{  950000, 0x00 }
};

static inline int
qm_tune(tuner, freq, lpf)
	struct pt3_tuner *tuner;
	u_int freq;
	int lpf;
{
	u_int sd, N, A;
	int i;
	u_char param[6];

	/* div2/vco_band */
	for (i = 0; i < 8; i++) {
		if (vco_table[i + 1].f <= freq && freq < vco_table[i].f)
			qm_write(tuner, 0x02, vco_table[i].val);
	}

	/* Calculate PLL parameters. */
	qm_pll_param(freq, &sd, &N, &A);

	param[0] = N;
	param[1] = 0x30 | (A & 0x0f);
	param[2] = 0x09;
	if (lpf)
		param[2] = 0x02;
	param[3] = sd >> 16;
	param[4] = sd >> 8;
	param[5] = sd;
	if (qm_write_n(tuner, 0x06, param, 6))
		return -1;

	if (!lpf) {
		if (qm_write(tuner, 0x13, 0x20))
			return -1;
	}

	if (lpf) {
		if (qm_write(tuner, 0x0c, 0x03))
			return -1;
		delay_ms(1);
		if (qm_write(tuner, 0x0c, 0xc3))
			return -1;
	} else {
		if (qm_write(tuner, 0x0c, 0x43))
			return -1;
		delay(1024);
		if (qm_write(tuner, 0x0c, 0xc3))
			return -1;
	}

	if (lpf)
		delay_ms(lpf_wait);
	else {
		if (tuner->fast_search)
			delay_ms(fast_search_wait);
		else
			delay_ms(normal_search_wait);
	}

	if (lpf) {
		if (qm_write(tuner, 0x08, 0x09) ||
		    qm_write(tuner, 0x13, 0x20))
			return -1;
	}

	return 0;
}

static inline u_int
s_channel_freq(channel)
	u_int channel;
{
	if (channel < 12) {
		/*
		 * BS-1 ... BS-23:
		 *   f = 1049.48 + 38.36 * channel;
		 */
		return 104948 + 3836 * channel;
	}
	if (channel < 24) {
		/* ND-2, ND-4, ..., ND-24 */
		return 159300 + 2000 * ((channel - 12) * 2 + 1);
	} else {
		/* ND-1, ND-3, ..., ND-23 */
		return 159300 + 2000 * ((channel - 24) * 2);
	}
}

static inline int
s_lock_status(tuner, locked)
	struct pt3_tuner *tuner;
	int *locked;
{
	int val;

	val = qm_read(tuner, 0x0d);
	if (val < 0)
		return -1;
	*locked = (val & 0x40) ? 1 : 0;
	return 0;
}

static int
s_tune(tuner, channel)
	struct pt3_tuner *tuner;
	u_int channel;
{
	u_int freq;
	int i, locked;

	if (s_disable_agc(tuner))
		return -1;

	freq = s_channel_freq(channel) * 10;	/* kHz */
	if (tuner->unit == 0)
		freq -= 500;
	else
		freq += 500;

	if (qm_tune(tuner, freq, 1))
		return -1;

	for (i = 0; i < 100; i++) {
		if (s_lock_status(tuner, &locked))
			return -1;
		if (locked)
			goto locked;
		delay(1000);
	}
	return -1;
locked:
	if (s_enable_agc(tuner))
		return -1;
	return 0;
}

static int
mxl_write(tuner, data, len)
	struct pt3_tuner *tuner;
	void *data;
	int len;
{
	struct pt3_softc *sc = tuner->softc;
	int error;
	u_char buf[32];

	if (len > 30)
		panic("mxl_write");

	mutex_enter(&sc->sc_i2c_lock);
	i2c_init(sc);

	buf[0] = THROUGH;
	buf[1] = tuner->tuner_addr;
	bcopy(data, buf + 2, len);
	i2c_send(sc, tuner->fe_addr, buf, len + 2);
	error = i2c_run(sc);
	mutex_exit(&sc->sc_i2c_lock);
	return error;
}

static int
mxl_read(tuner, reg, data)
	struct pt3_tuner *tuner;
	u_int reg;
	void *data;
{
	struct pt3_softc *sc = tuner->softc;
	int error;
	u_char buf[2];

	buf[0] = 0xfb;
	buf[1] = reg;
	if (mxl_write(tuner, buf, 2))
		return -1;

	mutex_enter(&sc->sc_i2c_lock);
	i2c_init(sc);

	buf[0] = THROUGH;
	buf[1] = tuner->tuner_addr | 1;
	i2c_send(sc, tuner->fe_addr, buf, 2);
	i2c_recv(sc, tuner->fe_addr, 1);
	error = i2c_run(sc);
	i2c_result(sc, data, 1);
	mutex_exit(&sc->sc_i2c_lock);
	return error;
}

static inline int
mxl_setreg(tuner, reg, value)
	struct pt3_tuner *tuner;
	u_int reg, value;
{
	u_char data[2];

	data[0] = reg;
	data[1] = value;
	return mxl_write(tuner, data, 2);
}

static struct {
	u_int freq;		/* center frequency */
	u_int off:16;
	u_int shf_val:8;	/* spur shift value */
	u_int shf_dir:8;	/* shift direction */
} spur_shifts[] = {
	{  64500, 500, 0x92, 0x07 },
	{ 191500, 300, 0xe2, 0x07 },
	{ 205500, 500, 0x2c, 0x04 },
	{ 212500, 500, 0x1e, 0x04 },
	{ 226500, 500, 0xd4, 0x07 },
	{  99143, 500, 0x9c, 0x07 },
	{ 173143, 500, 0xd4, 0x07 },
	{ 191143, 300, 0xd4, 0x07 },
	{ 207143, 500, 0xce, 0x07 },
	{ 225143, 500, 0xce, 0x07 },
	{ 243143, 500, 0xd4, 0x07 },
	{ 261143, 500, 0xd4, 0x07 },
	{ 291143, 500, 0xd4, 0x07 },
	{ 339143, 500, 0x2c, 0x04 },
	{ 117143, 500, 0x7a, 0x07 },
	{ 135143, 300, 0x7a, 0x07 },
	{ 153143, 500, 0x01, 0x07 }
};

static u_char idac_setting[] = {
	0x0d, 0x00,
	0x0c, 0x67,
	0x6f, 0x89,
	0x70, 0x0c,
	0x6f, 0x8a,
	0x70, 0x0e,
	0x6f, 0x8b,
	0x70, 0x10 + 12
};

static u_char tune_data[] = {
	0x13, 0x00,	/*  0 */
	0x3b, 0xc0,	/*  2 */
	0x3b, 0x80,	/*  4 */
	0x10, 0x95,	/*  6 */
	0x1a, 0x05,	/*  8 */
	0x61, 0x00,	/* 10 */
	0x62, 0xa0,	/* 12 */

	0x11, 0x40,	/* 14 */
	0x12, 0x0e,	/* 16 */
	0x13, 0x01	/* 18 */
};

static int
mxl_rftune(tuner, freq)
	struct pt3_tuner *tuner;
	u_int freq;
{
	u_int x;
	int i;
	u_char data[20];

	tuner->freq = freq;
	bcopy(tune_data, data, sizeof tune_data);

	/* Convert to 10.6 fixed point format in MHz. */
	x = (freq + 7812) / (1000000 / 64);
	data[15] = x;
	data[17] = x >> 8;

	for (i = 0; i < sizeof spur_shifts / sizeof spur_shifts[0]; i++) {
		int center = spur_shifts[i].freq * 1000;
		int offset = spur_shifts[i].off * 1000;

		if (abs((int)freq - center) <= offset) {
			data[11] = spur_shifts[i].shf_val;
			data[13] = 0xa0 | spur_shifts[i].shf_dir;
			break;
		}
	}

	if (mxl_write(tuner, data, 14))
		return -1;
	delay_ms(1);
	if (mxl_write(tuner, data + 14, 6))
		return -1;
	delay_ms(1);
	delay_ms(30);
	if (mxl_setreg(tuner, 0x1a, 0x0d))
		return -1;
	if (mxl_write(tuner, idac_setting, sizeof idac_setting))
		return -1;
	return 0;
}

static inline u_int
t_channel_freq(channel)
	u_int channel;
{
	u_int offset = 0;

	if (channel >= 12)
		offset += 2;
	if (channel >= 17)
		offset -= 2;
	if (channel >= 63)
		offset += 2;
	return 93 + channel * 6 + offset;
}

static inline int
t_lock_status(tuner, rfsynth, refsynth)
	struct pt3_tuner *tuner;
	int *rfsynth, *refsynth;
{
	u_char data;

	if (mxl_read(tuner, 0x16, &data)) {
		*rfsynth = *refsynth = 0;
		return -1;
	}
	*rfsynth  = ((data & 0x0c) == 0x0c) ? 1 : 0;
	*refsynth = ((data & 0x03) == 0x03) ? 1 : 0;
	return 0;
}

static int
t_tune(tuner, channel, offset)
	struct pt3_tuner *tuner;
	u_int channel;
	int offset;
{
	u_int base, freq;
	int i, locked1, locked2;

	if (t_disable_agc(tuner))
		return -1;

	base = t_channel_freq(channel);

	/* freq = (7 * base + 1 + offset) * 1000000 / 7 */
	freq = base * 1000000 + (1 + offset) * 1000000 / 7;

	if (mxl_rftune(tuner, freq))
		return -1;

	for (i = 0; i < 1000; i++) {
		t_lock_status(tuner, &locked1, &locked2);
		if (locked1 && locked2)
			goto locked;
		delay(1000);
	}
	return -1;

locked:
	if (t_enable_agc(tuner))
		return -1;
	return 0;
}

static inline int
slptim(tuner, sleep)
	struct pt3_tuner *tuner;
	int sleep;
{
	return fe_write1(tuner, 0x03, sleep ? 0x90 : 0x80);
}

static int
t_sleep(tuner)
	struct pt3_tuner *tuner;
{
	static u_char data[] = {
		0x01, 0x00,
		0x13, 0x00
	};

	if (t_disable_agc(tuner))
		return -1;
	if (mxl_write(tuner, data, 4))
		return -1;
	if (slptim(tuner, 1))
		return -1;
	return 0;
}

static int
t_wakeup(tuner)
	struct pt3_tuner *tuner;
{
	if (slptim(tuner, 0))
		return -1;
	if (mxl_setreg(tuner, 0x01, 0x01))
		return -1;
	if (mxl_rftune(tuner, tuner->freq))
		return -1;
	return 0;
}
